'''Map model.'''

import random

from concepts.geometry import Position, Rectangle, Direction



class Map(Rectangle):
    def __init__(self, model, width, height):
        Rectangle.__init__(self, 0, 0, width, height)
        
        self.model = model
        self.tiles = [[Tile(model, self, x, y)
                       for y in range(height)]
                       for x in range(width)]
        self.reverse = [[self.tiles[x][y]
                         for x in range(width)]
                         for y in range(height)]
        self.generate()
        
    def tile(self, pos):
        if (pos.x < 0 or pos.x >= self.width or
            pos.y < 0 or pos.y >= self.height):
            return None
        
        return self.tiles[pos.x][pos.y]
    
    def random_tile(self, acceptable = lambda tile: True):
        tiles = filter(acceptable, self)
        
        if len(tiles) == 0:
            return None
        
        return random.choice(tiles)
    
    def closest(self, source, acceptable = lambda tile: True):
        tiles = filter(acceptable, self)
        tiles.sort(lambda a, b: cmp(a.distance(source), b.distance(source)))
        
        if tiles:
            return tiles[0]
        else:
            return None

    def top_tile(self, prop, acceptable = lambda tile: True):
        tiles = filter(acceptable, self)
        tiles.sort(lambda a, b: cmp(prop(b), prop(a)))
        
        if tiles:
            return tiles[0]
        else:
            return None

    def start_tile(self):
        return self._start_tile
    
    def center_tile(self):
        return self.tile(Position(self.width / 2, self.height / 2))
    
    def tiles_in_radius(self, pos, radius, double_x = False):
        tiles = []
        
        for dx in range(-radius * (1 + double_x), radius * (1 + double_x) + 1):
            for dy in range(-radius, radius + 1):
                dir  = Direction(dx, dy)
                tile = self.tile(pos.in_direction(dir))
                
                if tile is not None:
                    tiles.append(tile)
        
        tiles.sort(lambda a, b: cmp(a.distance(pos), b.distance(pos)))

        return tiles

    def unbuilt_land_tiles(self):
        return filter(lambda tile:
                      not tile.is_built() and
                      tile.water_depth() == 0, self)

    def unbuilt_land_tile(self):
        return self.random_tile(lambda tile:
                                not tile.is_built() and
                                tile.water_depth() == 0)
        
    def __getitem__(self, nr):
        x = nr % self.width
        y = nr / self.width
        
        return self.tile(Position(x, y))
    
    def __len__(self):
        return self.area()
    
    def __iter__(self):
        return MapIterator(self)
        
    def generate(self):
        raise 'Not implemented'
    
    
    
class MapIterator:
    def __init__(self, map):
        self.map = map
        self.nr = 0
        
    def next(self):
        if self.nr >= len(self.map):
            raise StopIteration
        
        item = self.map[self.nr]
        self.nr += 1
        
        return item
        
    def __iter__(self):
        return self



class Tile(Position):
    def __init__(self, model, map, x, y):
        Position.__init__(self, x, y)
        
        self.model = model
        self.map = map
        self.terrain = None
        self._structure = None
    
    def destroy_buildings(self):
        self._structure = None
        
    def become(self, kind, *params):
        self.terrain = kind(self.model, self, *params)
    
    def build(self, kind, *params):
        self._structure = kind(self.model, self, *params)
    
    def is_built(self):
        return (self.structure() is not None)
    
    def closest(self, acceptable = lambda tile: True):
        return self.map.closest(self, acceptable)
    
    def current_terrain(self):
        return self.terrain
    
    def structure(self):
        return self._structure
    
    def sub_location(self, avatar = None):
        if avatar is None:
            avatar = self.model.current_avatar()
        
        if self.structure() is not None and (avatar.is_inside() or
                                             self.structure().is_outside()):
            return self.structure()
        
        if self.terrain is not None and not avatar.is_inside():
            return self.terrain
        
        return None
    
    def livability(self):
        if self.terrain is None:
            return 0
        
        return self.terrain.life()
    
    def water_depth(self):
        if self.terrain is None:
            return 0
        
        return self.terrain.water_depth()
    
    def can_move_in(self, avatar):
        if self.terrain is not None:
            return self.terrain.can_move_in(avatar)
        
        return False
    
    def move_in(self, avatar):
        if not self.can_move_in(avatar):
            return
        
        if self.terrain is not None:
            self.terrain.move_in(avatar)
    
    def events(self, avatar):
        return self.sub_location(avatar).events(avatar)
    
    def situation(self):
        return self.sub_location().situation()
    
    def set_situation(self, situation):
        self.sub_location().set_situation(situation)
        
    def resolve_situation(self, situation):
        self.sub_location().resolve_situation(situation)
    
    def look_around(self):
        return self.sub_location().look_around()



class TileEnhancement:
    def __init__(self, model, tile, name):
        self._model = model
        self._tile = tile
        self._name = name
    
    def model(self):
        return self._model
    
    def tile(self):
        return self._tile
    
    def name(self):
        return self._name
    
    def can_move_in(self, avatar):
        return True
    
    def move_in(self, avatar):
        pass
    
    def events(self, avatar):
        return []
    
    def situation(self):
        return None
    
    def resolve_situation(self, situation):
        pass
    
    def set_situation(self, situation):
        pass
    
    def __str__(self):
        return self.name()

    def look_around(self):
        raise 'Not implemented'
    


class Terrain(TileEnhancement):
    def livability(self):
        raise 'Not implemented'


    
class Structure(TileEnhancement):
    def __init__(self, model, tile, hidden, outside, name):
        TileEnhancement.__init__(self, model, tile, name)
        
        self._hidden  = hidden
        self._outside = outside
        
    def is_hidden(self):
        return self._hidden
        
    def is_outside(self):
        return self._outside
    
    def description(self):
        return []
    
    def unhide(self):
        self._hidden = False
