'''Actions in the Phoenix game.'''

from concepts.actions import Action

from phoenix.events import *



class MoveAction(Action):
    def __init__(self, situation, dir):
        Action.__init__(self, situation, 'Move', str(dir))
        
        self.dir = dir
    
    def can_do(self):
        return self.model().current_avatar().can_move(self.dir)
    
    def do(self):
        self.model().current_avatar().move(self.dir)



class HealAction(Action):
    def __init__(self, situation, min, max):
        Action.__init__(self, situation, 'Heal Yourself')
        
        self._min = min
        self._max = max
    
    def min(self):
        return self._min()
    
    def max(self):
        return self._max()
    
    def can_do(self):
        return (self.model().current_avatar().gems('water') >= 1 and
                not self.model().current_avatar().current_tile().is_built())
    
    def do(self):
        self.model().add_events([Heal(self, self.model().current_avatar(),
                                      self.min(), self.max())])
        self.model().current_avatar().consume_gem('water')



class TradeGemAction(Action):
    def __init__(self, situation, domain, chance):
        Action.__init__(self, situation, 'Trade Gems')
        
        self._domain = domain
        self._chance = chance()
    
    def domain(self):
        return self._domain
    
    def chance(self):
        return self._chance
    
    def can_do(self):
        tile = self.model().current_avatar().current_tile()
        
        return (self.model().current_avatar().gems(self.domain()) >= 1 and
                not tile.is_built() and
                len(tile.terrain.domains()) > 0)
    
    def do(self):
        self.model().add_events([TradeGems(self, self.model().current_avatar(),
                                           self.domain(), self.chance())])



class BlackMarketAction(Action):
    def __init__(self, situation, trade):
        Action.__init__(self, situation, str(trade))
        
        self._trade = trade
    
    def trade(self):
        return self._trade
    
    def info(self):
        return '%dg' % self.trade().gold()
    
    def always_known(self):
        return True
    
    def can_do(self):
        trade  = self.trade()
        avatar = self.model().current_avatar()
        
        if trade.is_sale():
            return avatar.gems(trade.domain()) >= trade.num()
        else:
            return avatar.gold() >= trade.gold()
    
    def do(self):
        trade  = self.trade()
        avatar = self.model().current_avatar()
        
        trade.market().remove_trade(trade)
        
        if trade.is_sale():
            avatar.add_gold(trade.gold())
        else:
            avatar.lose_gold(trade.gold())
            
        for nr in range(trade.num()):
            if trade.is_sale():
                avatar.consume_gem(trade.domain())
            else:
                avatar.add_gem(trade.domain())



class AlchemistAction(Action):
    def __init__(self, situation, potion, market):
        Action.__init__(self, situation, 'Potion "H%d"' % potion[0])
        
        self._potion = potion
        self._market = market
    
    def potion(self):
        return self._potion
    
    def market(self):
        return self._market
    
    def heal(self):
        return self.potion()[0]
    
    def gold(self):
        return self.potion()[1]
    
    def info(self):
        return '%dg' % self.gold()
    
    def always_known(self):
        return True
    
    def can_do(self):
        return self.model().current_avatar().gold() >= self.gold()
    
    def do(self):
        avatar = self.model().current_avatar()
        
        self.market().buy_potion()
        
        avatar.lose_gold(self.gold())
        avatar.heal(self.heal())



class DirectAction(Action):
    def __init__(self, name, cost, text, happenings):
        Action.__init__(self, None, name)
        
        self._cost       = cost
        self._text       = text
        self._happenings = happenings
        
    def cost(self):
        return self._cost
    
    def text(self):
        return self._text
    
    def happenings(self):
        return self._happenings
    
    def info(self):
        if self.cost() is None:
            return ''
        
        return '%dg' % self.cost()
    
    def always_known(self):
        return True
    
    def can_do(self):
        if self.cost() is None:
            return True
        
        return self.model().current_avatar().gold() >= self.cost()
    
    def do(self):
        if self.cost() is not None:
            self.model().current_avatar().lose_gold(self.cost())
            
        self.model().add_events([DirectEvent(self.text(), self.happenings())])
        


class TownAction(Action):
    def __init__(self, situation, item, market):
        Action.__init__(self, situation, 'Buy: %s' % item[0])
        
        self._item   = item
        self._market = market
    
    def item(self):
        return self._item
    
    def market(self):
        return self._market
    
    def name(self):
        return self.item()[0]
    
    def places(self):
        return self.item()[1]
    
    def gold(self):
        return self.item()[2]
    
    def info(self):
        return '%dg' % self.gold()
    
    def always_known(self):
        return True
    
    def can_do(self):
        return self.model().current_avatar().gold() >= self.gold()
    
    def do(self):
        avatar = self.model().current_avatar()
        map    = self.model().current_map()
        
        self.market().buy_map()
        
        avatar.lose_gold(self.gold())
        tile = map.closest_tile(lambda tile: tile.is_built() and
                                tile.structure().is_hidden() and
                                reduce(lambda a, type: 
                                       a or isinstance(tile.structure(), type), 
                                       self.places(), False), avatar)
        
        text = ['You buy the map.']
        
        if tile is None:
            text += ['It reveals nothing new.']
            event = []
        else:
            dir   = avatar.direction_to(tile).compass_name()
            text += ['It reveals %s to the %s.' % 
                     (str(tile.structure()).lower(), dir)]
            event = [lambda: tile.structure().unhide()]
            
        self.model().add_events([DirectEvent(text, event)])



class FightAction(Action):
    def __init__(self, situation, battle = None):            
        Action.__init__(self, situation, 'Fight')

        if battle is None:
            battle = situation
        
        self._battle = battle
        
    def battle(self):
        return self._battle
    
    def can_do(self):
        return True
    
    def info(self):
        return self.battle().combat_info()
        
    def do(self):
        self.battle().fight()



class BlastAction(Action):
    def __init__(self, situation, chance):            
        Action.__init__(self, situation, 'Blast')

        self._chance = chance
        
    def battle(self):
        return self.situation
    
    def chance(self):
        return self._chance()
    
    def can_do(self):
        return self.model().current_avatar().gems('fire') > 0
    
    def info(self):
        return '100%% (%d%%)' % self.chance()
        
    def do(self):
        self.battle().blast(self.chance())



class ScorchAction(Action):
    def __init__(self, situation, degree):
        Action.__init__(self, situation, 'Scorch Earth')

        self._degree = degree
    
    def degree(self):
        return self._degree()
    
    def can_do(self):
        return (self.model().current_avatar().gems('fire') >= 1 and
                not self.model().current_avatar().current_tile().is_built())
        
    def do(self):
        self.model().add_events([ScorchTheEarth(self.model().current_avatar(), 
                                                self.model(), self.degree())])




class BribeAction(Action):
    def __init__(self, situation, battle = None):            
        Action.__init__(self, situation, 'Bribe')

        if battle is None:
            battle = situation
        
        self._battle = battle
        
    def battle(self):
        return self._battle
    
    def can_do(self):
        return self.battle().avatar().gold() >= self.battle().monster().greed()
    
    def info(self):
        return self.battle().bribe_info()
        
    def do(self):
        self.battle().bribe()



class RunAction(Action):
    def __init__(self, situation, battle = None):            
        Action.__init__(self, situation, 'Run')

        if battle is None:
            battle = situation
        
        self._battle = battle
        
    def battle(self):
        return self._battle
    
    def can_do(self):
        return True
    
    def info(self):
        return self.battle().run_info()
    
    def do(self):
        self.battle().run()



class GainSkillAction(Action):
    def __init__(self, situation, skill):
        Action.__init__(self, situation, 'Gain Skill', str(skill))
        
        self._skill = skill()
    
    def skill(self):
        return self._skill
    
    def can_do(self):
        avatar = self.model().current_avatar() 
        domain = self.skill().domain()
        
        return (avatar.feathers(domain) >= self.skill().required_feathers() and
                not self.skill() in avatar.skills(domain))
    
    def do(self):
        self.model().current_avatar().add_skill(self.skill())
        self.model().current_situation().resolve()



class RefuseSkillAction(Action):
    def __init__(self, situation):
        Action.__init__(self, situation, 'Refuse the Skill')
    
    def can_do(self):
        return True
    
    def do(self):
        self.model().current_situation().resolve()

