'''UI parts specific to the phoenix game.'''

import math
import random

from ui.cursesemu import KEY_UP, KEY_DOWN, KEY_LEFT, KEY_RIGHT

from concepts.geometry import Position, Rectangle

from ui.cursed import CursesGameUI
from ui.properties import *
from ui.windows import ActionWindow, MapWindow, Window



class PhoenixUI(CursesGameUI):
    KEYS = {'Quit'                : ord('Q'),
            'Do Nothing'          : ord(' '),
            'Fight'               : ord('f'),
            'Run'                 : ord('r'),
            'Bribe'               : ord('b'),
            'Blast'               : ord('l'),
            'Refuse the Skill'    : ord('r'),
            'Heal Yourself'       : ord('h'),
            'Trade Gems'          : ord('t'),
            'Take some Gold'      : ord('t'),
            'Scorch Earth'        : ord('s'),
            'Inspect the Mine'    : ord('i'),
            'Search the Mine'     : ord('s'),
            'Step on the Platform': ord('s'),
            'Dig'                 : ord('d'),
            'Open'                : ord('o'),
            'Leave'               : ord('l'),
            'Invoke the Rainbow'  : ord('r'),
            'Ascend the Rainbow'  : ord('a'),
            'Accept your Fate'    : ord(' '),
            'Move: North'         : KEY_UP,
            'Move: South'         : KEY_DOWN,
            'Move: West'          : KEY_LEFT,
            'Move: East'          : KEY_RIGHT}
    
    DESCS = {ord(' '): 'Spc'}
    
    GROUPS = {'Move'      : ('Arr', 'Move'),
              'Gain Skill': ('Num', 'Gain Skill')}
    
    def init(self):
        self._key_nr = 0
        self._keys   = {}
        self.windows = {'action': self.create_window(ActionWindow, 
                                                     Rectangle(51, 16, 29,  8)),
                        'status': self.create_window(StatusWindow,
                                                     Rectangle(51,  0, 29,  3)),
                        'skill' : self.create_window(SkillWindow,
                                                     Rectangle(51,  3, 29, 13)),
                        'map'   : self.create_window(PhoenixMainWindow,
                                                     Rectangle( 0,  0, 51, 17)),
                        'text'  : self.create_window(MessageWindow,
                                                     Rectangle( 0, 17, 51,  7))}
        
    def prepare_turn(self):
        self._key_nr = 0
        self._keys   = {}
        
    def window(self, name):
        return self.windows.get(name, None)

    def group_for(self, action):
        parts = str(action).split(':')
        
        return self.GROUPS.get(parts[0], None)
    
    def char_for(self, action):
        key = self.key_for(action)
        
        if key is None:
            return None
        
        if key < 256 and not self.DESCS.has_key(key):
            return chr(key)
        
        return self.DESCS.get(key, None)
    
    def key_for(self, action):
        if not self.KEYS.has_key(str(action)):
            if self._keys.has_key(str(action)):
                return self._keys[str(action)]
            
            if self._key_nr >= 9:
                return None
            
            self._key_nr += 1
            self._keys[str(action)] = ord(str(self._key_nr))
            
            return ord(str(self._key_nr))
        
        return self.KEYS[str(action)]
    
    def displayed_windows(self):
        return self.windows.values()

    def delay(self, ms = 500):
        self.stdscr.refresh()
        curses.delay_output(ms)

    def paint_progress(self, y, x, width, char, attrs, num, div):
        cx = x + len(str(num)) + 1
        cy = y
        cn = 0
        cw = width - len(str(num)) - 1
        
        for nr in range(1, num / div + 1 + (num % div > 0)):
            cn = nr * div
            
            if cn > num:
                cn = num
                
            self.stdscr.addstr(y, x, '%*d' % (len(str(num)), cn), attrs)
            self.stdscr.addstr(cy, cx, char, attrs)
            
            cw -= 1
            cx += 1
            
            if cw <= 0:
                cw  = width
                cx  = x
                cy += 1
                
            self.stdscr.refresh()
            self.delay(30)

    def do_scores(self):
        scr  = self.stdscr
        attr = lambda attributes: reduce(lambda a, b: a | b, attributes)
        
        (score, stuff) = self.model.score()
        
        if self.model.has_won():
            caption = 'You escaped the Isle!'
        else:
            caption = 'You died.'
        
        scr.erase()
        scr.box()
        
        scr.addstr(1, 1, caption.center(78), attr([WHITE]))
        self.delay()
        
        scr.addstr(5, 12, 'Gold', attr([WHITE]))
        scr.addstr(5, 62, 'Points', attr([WHITE]))
        self.delay()
        
        self.paint_progress(6, 12, 41, '$', attr([YELLOW]), stuff['gold'], 2500)
        
        scr.addstr(6, 58, '%10d' % stuff['gold'])
        self.delay()
        
        scr.addstr(9, 12, 'Gems', attr([WHITE]))
        self.delay()
        
        gems     = stuff['gems']
        feathers = stuff['feathers']
        
        if gems['water'] > 0:
            self.paint_progress(10, 12, 8, '*', attr([BLUE]),   gems['water'], 10)
            self.delay()
        
        if gems['earth'] > 0:
            self.paint_progress(10, 23, 8, '*', attr([GREEN]),  gems['earth'], 10)
            self.delay()
        
        if gems['fire'] > 0:
            self.paint_progress(10, 34, 8, '*', attr([RED]),    gems['fire'],  10)
            self.delay()

        if gems['air'] > 0:
            self.paint_progress(10, 45, 8, '*', attr([YELLOW]), gems['air'],   10)
            self.delay()
        
        scr.addstr(10, 58, '%10d' % (50 * reduce(lambda x, n: x + n, stuff['gems'].values(), 0)))
        self.delay()
        
        if self.model.has_won():
            scr.addstr(13, 12, 'Feathers', attr([WHITE]))
            self.delay()
        
            if feathers['water'] > 0:
                self.paint_progress(14, 12, 8, '/', attr([BLUE]),   feathers['water'], 1)
                self.delay()
            
            if feathers['earth'] > 0:
                self.paint_progress(14, 23, 8, '/', attr([GREEN]),  feathers['earth'], 1)
                self.delay()
                
            if feathers['fire'] > 0:
                self.paint_progress(14, 34, 8, '/', attr([RED]),    feathers['fire'],  1)
                self.delay()
        
            if feathers['air'] > 0:
                self.paint_progress(14, 45, 8, '/', attr([YELLOW]), feathers['air'],   1)
                self.delay()
        
            scr.addstr(14, 58, '%10d' % (1000 * int(math.sqrt(reduce(lambda x, n: x * n, stuff['feathers'].values(), 1)))))
            self.delay()
            
        scr.addstr(18, 12, 'Total', attr([WHITE]))
        scr.addstr(18, 58, '%10d' % score, attr([WHITE]))
        
        scr.refresh()
        curses.flushinp()
        scr.getch()
        
        scr.erase()
        scr.box()
        
        scr.addstr(1, 1, 'Highscores'.center(78), attr([WHITE]))
        self.delay()
        
        scr.addstr(3,  2, 'Place', attr([WHITE]))
        scr.addstr(3, 11, 'Date', attr([WHITE]))
        scr.addstr(3, 23, 'Gold', attr([WHITE]))
        scr.addstr(3, 40, 'Gems', attr([WHITE]))
        scr.addstr(3, 56, 'Feathers', attr([WHITE]))
        scr.addstr(3, 72, 'Points', attr([WHITE]))
        self.delay()
        
        place = 0
        
        for x in self.model.highscores().top_scores(19):
            place += 1
            
            (date, x) = x
            (score, stuff) = x
            
            scr.addstr(3 + place,  2, '%4d.' % place)
            scr.addstr(3 + place,  8, '%s' % str(date))
            scr.addstr(3 + place, 20, '[', attr([WHITE]))
            scr.addstr(3 + place, 21, '%8d' % stuff['gold'], attr([YELLOW]))
            scr.addstr(3 + place, 29, ']  [', attr([WHITE]))
            scr.addstr(3 + place, 33, '%3d'% stuff['gems']['water'], attr([BLUE]))
            scr.addstr(3 + place, 38, '%3d'% stuff['gems']['earth'], attr([GREEN]))
            scr.addstr(3 + place, 43, '%3d'% stuff['gems']['fire'], attr([RED]))
            scr.addstr(3 + place, 48, '%3d'% stuff['gems']['air'], attr([YELLOW]))
            scr.addstr(3 + place, 51, ']  [', attr([WHITE]))
            scr.addstr(3 + place, 55, '%d'% stuff['feathers']['water'], attr([BLUE]))
            scr.addstr(3 + place, 58, '%d'% stuff['feathers']['earth'], attr([GREEN]))
            scr.addstr(3 + place, 61, '%d'% stuff['feathers']['fire'], attr([RED]))
            scr.addstr(3 + place, 64, '%d'% stuff['feathers']['air'], attr([YELLOW]))
            scr.addstr(3 + place, 65, ']', attr([WHITE]))
            scr.addstr(3 + place, 68, '%10d'% score)
            self.delay(100)
            
        scr.refresh()
        curses.flushinp()
        scr.getch()

        

class SkillWindow(Window):
    DOMAIN_COLORS = [('water', BLUE),
                     ('earth', GREEN),
                     ('fire',  RED),
                     ('air',   YELLOW)]
    
    def display(self):
        self.clear()
        
        nr = 0
        lines = []
        avatar = self.model.current_avatar()
        
        for domain in self.DOMAIN_COLORS:
            (name, color) = domain
                        
            gems   = avatar.gems(name)
            g_s    = ''
            g_text = ''
            
            feathers = avatar.feathers(name)
            f_s      = ''
            f_text   = ''
            
            if gems != 1:
                g_s = 's'
            
            if feathers != 1:
                f_s = 's'
            
            if gems > 0:
                g_text = '%3d gem%1s' % (gems, g_s)
                
            if feathers > 0:
                f_text = '%d feather%1s' % (feathers, f_s)
                
            text = ('%-5s   %10s %8s' %
                    (name.capitalize(), f_text, g_text))
            
            if gems + feathers > 0:
                lines += [(text, color)]
            
            for skill in avatar.skills(name):
                text = ' %-18s %7s' % (str(skill), skill.info())
                lines += [(text, GREY)]
                
        for line in lines:
            (text, color) = line
            
            self.write(Position(0, nr), text, [color])
            nr += 1



class MessageWindow(Window):
    def display(self):
        self.clear()
        
        tile   = self.model.current_avatar().current_tile()
        color  = self.ui.window('map').color_for_tile(tile)
        events = self.model.current_avatar().events()
        
        self.write(Position(0, 0), tile.look_around(), [color])
        
        line = 1
        
        for event in events:
            if event.is_flavor():
                color = BLACK
            else:
                color = GREY
            
            for text in event.description():
                self.write(Position(0, line), text, [color])
                line += 1
        
        
        
class PhoenixMapWindow(MapWindow):
    LIFE_COLORS  = [BRICK, BLACK, YELLOW, GREEN, GRASS]
    WATER_COLORS = [BLUE, MARINE]

    DOMAIN_COLORS = {'water': BLUE,
                     'earth': GREEN,
                     'fire' : RED,
                     'air'  : YELLOW}
    
    UNIT_COLORS = {'Avatar'          : WHITE,
                   'Dead Avatar'     : BLACK,
                   'Destroyed Avatar': BRICK}
    
    CHARS = {'Avatar'            : '@',
             'Dead Avatar'       : '@',
             'Destroyed Avatar'  : '#',
             'Plains'            : '.',
             'Lake'              : '~',
             'Forest'            : '&',
             'Hills'             : ';',
             'Mountain'          : '^',
             'Glacier'           : '^',
             'Rainbow Mountain'  : '^',
             'Rainbow Foundation': '^',
             'Lava'              : '~',
             'Apocalyptic Place' : '#',
             'Phoenix Nest'      : '_',
             'Sunken City'       : '*',
             'Tree of Ages'      : '#',
             'Lava Forge'        : '%',
             'Ivory Tower'       : '!',
             'Fountain of Youth' : '=',
             'Dragon Cave'       : '=',
             'Wishing Well'      : '=',
             'Crystal Platform'  : '=',
             'Town'              : '*',
             'Mine'              : '>',
             'Black Market'      : '+',
             'Alchemist\'s Hut'  : '+',
             'Small Hut'         : '+'}
    
    def char_for_tile(self, tile):
        if tile.is_built() and not tile.structure().is_hidden():
            return self.CHARS.get(tile.structure().name(), None)
            
        return self.CHARS.get(str(tile.current_terrain()), None)
    
    def color_for_tile(self, tile):
        if str(tile.terrain) == 'Apocalyptic Place':
            if random.randint(0, 3):
                return BRICK
            else:
                return RED
            
        if str(tile.terrain) == 'Rainbow Mountain':
            return random.choice(self.DOMAIN_COLORS.values())
        
        if tile.is_built() and not tile.structure().is_hidden():
            if tile.structure().domain() is not None:
                return self.DOMAIN_COLORS[tile.structure().domain()]
            elif tile.structure().is_dark_place():
                return BLACK
            elif tile.structure().is_wooden():
                return BROWN
            else:
                return WHITE
        
        if str(tile.terrain) == 'Mountain':
            return GREY

        if str(tile.terrain) == 'Glacier':
            return WHITE
        
        if tile.water_depth() > 0:
            nr = tile.water_depth() * len(self.WATER_COLORS) / 101
            
            return self.WATER_COLORS[nr]
        
        nr = tile.livability() * len(self.LIFE_COLORS) / 101
        
        return self.LIFE_COLORS[nr]
    
    def char_for_unit(self, unit):
        return self.CHARS.get(str(unit), None)
    
    def color_for_unit(self, unit):
        return self.UNIT_COLORS.get(str(unit), None)
    
    def map(self):
        return self.model.current_map()
    
    def units(self):
        return [self.model.current_avatar()]
    
    def center(self):
        return self.model.current_avatar().position()
            
        
    
class PhoenixMainWindow(PhoenixMapWindow):
    def display(self):
        if not self.model.is_skill_selection():
            PhoenixMapWindow.display(self)
            return
        
        self.clear()
        
        avatar    = self.model.current_avatar()
        situation = self.model.current_situation()
        domain    = situation.domain()
        color     = self.DOMAIN_COLORS[domain]
        skills    = self.model.skills(domain)
        feathers  = avatar.feathers(domain)
        an        = 'a'
        
        if domain[0] in 'aeiou':
            an = 'an'
            
        self.write(Position(0, 0),
                   'Select %s %s magic skill' % (an, domain), [WHITE])
        self.write(Position(0, 2), 
                   'The feather\'s magic lets you gain a skill:', [GREY])
        
        skill_nr = 0
        
        for skill in skills:
            skill_nr += 1
            line      = 4 * ((skill_nr + 1) / 2)
            column    = ((skill_nr - 1) % 2) * self.rect.width / 2
            cur_color = color
            low_color = GREY
            
            if skill() in avatar.skills(domain):
                cur_color = BLACK
                low_color = BLACK

                self.write(Position(column, line + 1),
                           '[Already taken]', [cur_color])
            
            if feathers < skill().required_feathers():
                cur_color = BLACK
                low_color = BLACK
                
                self.write(Position(column, line + 1),
                           '[Needs %d feathers]' % skill().required_feathers(),
                           [cur_color])
                                
            self.write(Position(column, line),
                       '%d) %s' % (skill_nr, str(skill())), [cur_color])
            self.write(Position(column, line + 2),
                       skill().description(), [low_color])
        
    
               
class StatusWindow(Window):
    def display(self):
        self.clear()
        
        health = self.model.current_avatar().health()
        gold   = self.model.current_avatar().gold()
        
        if health < 100:
            color = RED
        else:
            color = WHITE
        
        self.write(Position( 0, 0), 'Health:',            [GREY])
        self.write(Position( 7, 0), str(health).rjust(5), [color])
        self.write(Position(15, 0), 'Gold:',              [GREY])
        self.write(Position(21, 0), str(gold).rjust(6),   [WHITE])
        