'''Events in the Phoenix game.'''

import random

from concepts.events import Event

from phoenix.structures import *

import phoenix.situations

def Battle(*foo):
    return phoenix.situations.Battle(*foo)

def DirectSituation(*foo):
    return phoenix.situations.DirectSituation(*foo)



class Flavor(Event):
    def __init__(self, source, target, text):
        Event.__init__(self, source, target)
        
        self._description = random.choice(text)
    
    def is_flavor(self):
        return True
    
    def description(self):
        return [self._description]
    
    def do(self):
        pass



class DirectEvent(Event):
    def __init__(self, text, happenings):
        Event.__init__(self, None, None)
        
        self._description = text
        self._happenings  = happenings
    
    def description(self):
        return self._description
    
    def happenings(self):
        return self._happenings
    
    def do(self):
        for happening in self.happenings():
            happening()



class TextEvent(Event):
    def __init__(self, text):
        Event.__init__(self, None, None)
        
        self._description = text
    
    def description(self):
        return self._description
    
    def do(self):
        pass            



class PhoenixFlight(Event):
    def __init__(self, source, target, nest_tile):
        Event.__init__(self, source, target)
        
        self._nest_tile = nest_tile
        self._was_home  = None
        
    def tile(self):
        return self._nest_tile
    
    def nest(self):
        if self.tile() is None:
            return None
        
        return self.tile().structure()
    
    def was_home(self):
        return self._was_home
    
    def see_direction(self):
        return False
    
    def description(self):
        if self.nest() is None:
            return ['You stare into the air.',
                    'You see nothing but empty sky.']
        
        name = self.nest().domain()
        dir  = self.source().tile().direction_to(self.tile())
        
        if self.was_home():
            dir.invert()
            
        dir_name = dir.compass_name()
        
        if name[0] in 'aeiou':
            messages = ['You see an %s phoenix flying %s!' % (name, dir_name)]
        else:
            messages = ['You see a %s phoenix flying %s!' % (name, dir_name)]
            
        if self.see_direction():
            if self.was_home():
                messages += ['It appears to head away from its nest.']
            else:
                messages += ['It appears to head towards its nest.']
                
        return messages
    
    def do(self):
        if self.nest() is None:
            return
        
        self._was_home = self.nest().is_home()
        self.nest().fly()
        


class Ambush(Event):
    def __init__(self, source, target, monster, peace):
        Event.__init__(self, source, target)
        
        self._monster = monster
        self._peace   = peace
    
    def monster(self):
        return self._monster
    
    def peace(self):
        return self._peace
    
    def description(self):
        name  = str(self.monster()).lower()
        add   = ''
        
        if self.peace():
            add = (' %s suddenly stops.' %
                   self.monster().gender().nominative().capitalize())
        elif self.target().fiery_aura() >= self.monster().strength():
            add = (' Your fiery aura kills %s!' %
                   self.monster().gender().accusative())
        
        if name[0] in 'aeiou':
            return ['An %s ambushes you!%s' % (name, add)]
        else:
            return ['A %s ambushes you!%s' % (name, add)]
    
    def do(self):
        if (not self.peace() and
            self.target().fiery_aura() < self.monster().strength()):
            self.source().set_situation(Battle(self.model(), 
                                               self.target(), self.monster()))
        


class SituationEvent(Event):
    def __init__(self, source, description, actions, move = True):
        Event.__init__(self, source, source.model().current_avatar())
        
        self._description = description
        self._actions     = actions
        self._move        = move
        
    def description(self):
        return self._description
    
    def actions(self):
        return self._actions
    
    def move(self):
        return self._move
    
    def model(self):
        return self.source().model()
    
    def do(self):
        situation = DirectSituation(self.model(), self.actions(), self.move())

        self.source().set_situation(situation)


    
class Trap(Event):
    def __init__(self, source, target, min, max):
        Event.__init__(self, source, target)
        
        self._damage = random.randint(min, max)
        self._evaded = target.is_evading_traps()
        
    def damage(self):
        return self._damage
    
    def resistance(self):
        return self.target().resistance(self.damage())

    def evaded(self):
        return self._evaded
    
    def description(self):
        if self.evaded():
            return ['You manage to evade a trap.']
            
        return [('You fall into a trap! You suffer %d damage.' % 
                 (self.damage() - self.resistance()))]
    
    def do(self):
        if not self.evaded():
            self.target().damage(self.damage())


    
class Heat(Event):
    def __init__(self, source, target, min, max):
        Event.__init__(self, source, target)
        
        self._damage = random.randint(min, max)
    
    def damage(self):
        return self._damage
    
    def resistance(self):
        return self.target().resistance(self.damage(), True)

    def description(self):
        return [('The sun burns down on you. You suffer %d damage.' % 
                 (self.damage() - self.resistance()))]
    
    def do(self):
        self.target().damage(self.damage(), True)


    
class LavaHeat(Event):
    def __init__(self, source, target, min, max):
        Event.__init__(self, source, target)
        
        self._damage = random.randint(min, max)
    
    def damage(self):
        return self._damage
    
    def resistance(self):
        return self.target().resistance(self.damage(), True)

    def description(self):
        return [('This place burns you! You suffer %d damage.' % 
                 (self.damage() - self.resistance()))]
    
    def do(self):
        self.target().damage(self.damage(), True)


    
class Drown(Event):
    def __init__(self, source, target, min, max):
        Event.__init__(self, source, target)
        
        self._damage = random.randint(min, max)
    
    def damage(self):
        return self._damage
    
    def resistance(self):
        return self.target().resistance(self.damage(), True)

    def description(self):
        return [('You drown! You suffer %d damage.' % 
                 (self.damage() - self.resistance()))]
    
    def do(self):
        self.target().damage(self.damage(), True)


    
class Heal(Event):
    def __init__(self, source, target, min, max):
        Event.__init__(self, source, target)
        
        self._heal = random.randint(min, max)
    
    def heal(self):
        return self._heal

    def description(self):
        return ['You concentrate on your wounds.',
                'You gain %d health.' % 
                self.heal()]
    
    def do(self):
        self.target().heal(self.heal())


    
class TradeGems(Event):
    def __init__(self, source, target, domain, chance):
        Event.__init__(self, source, target)
        
        target_domain = None
        terrain       = target.current_tile().terrain
        
        while target_domain is None or target_domain == domain:
            target_domain = random.choice(terrain.domains() * 3 +
                                          ['water', 'fire', 'earth', 'air'])

        num = 1
        
        if random.randint(1, 100) <= chance:
            num += 1
        
        if num > 1:
            self._gems = Gems(target.model(), target_domain, num)
        else:
            self._gems = Gem(target.model(), target_domain)
            
        self._domain = domain
        
    def domain(self):
        return self._domain
    
    def gems(self):
        return self._gems

    def description(self):
        item = str(self.gems())
        
        return (['You trade for %s.' % str(self.gems())] +
                self.gems().find_description())
    
    def do(self):
        self.target().consume_gem(self.domain())
        self.gems().found_by(self.target())
        

    
class FindItem(Event):
    def __init__(self, source, target, item):
        Event.__init__(self, source, target)
        
        self._item = item
    
    def item(self):
        return self._item
        
    def description(self):
        return (['You find %s.' % str(self.item())] + 
                self.item().find_description())
    
    def do(self):
        self.item().found_by(self.target())
        

    
class ConjureItem(Event):
    def __init__(self, source, target, item):
        Event.__init__(self, source, target)
        
        self._item = item
    
    def item(self):
        return self._item
        
    def description(self):
        return (['You conjure %s from thin air.' % str(self.item())] + 
                self.item().find_description())
    
    def do(self):
        self.item().found_by(self.target())

        
    
class ScorchTheEarth(Event):
    def __init__(self, source, target, degree):
        Event.__init__(self, source, target)
        
        self._degree = degree
    
    def degree(self):
        return self._degree
        
    def description(self):
        return ['You scorch the earth!']
    
    def do(self):
        self.model().current_avatar().scorch(self.degree())
        self.source().consume_gem('fire')


    
class FindStructure(Event):
    def __init__(self, source, target, structure):
        Event.__init__(self, source, target)
        
        self._structure = structure
        self._item      = None
        
    def structure(self):
        return self._structure
    
    def item(self):
        return self._item
    
    def description(self):
        structure = str(self.structure()).lower()
        
        if self.item():
            desc = (self.structure().description() + 
                    self.item().find_description())
            text = 'You find %s containing %s.' % (structure, str(self.item()))
        else:
            desc = self.structure().description()
            text = 'You find %s.' % structure
        
        return [text] + desc
    
    def do(self):
        self.structure().unhide()
        
        self._item = self.structure().get_contents()
        
        if self._item is not None:
            self.item().found_by(self.target())


    
class SeeStructure(Event):
    def __init__(self, source, target, structure):
        Event.__init__(self, source, target)
        
        self._structure = structure
        self._item      = None
    
    def structure(self):
        return self._structure
    
    def item(self):
        return self._item
        
    def description(self):
        structure = str(self.structure()).lower()
        
        if self.item():
            desc = (self.structure().description() + 
                    self.item().find_description())
            text = 'You see %s containing %s.' % (structure, str(self.item()))
        else:
            desc = self.structure().description()
            text = 'You see %s.' % structure
        
        return [text] + desc
    
    def do(self):        
        self._item = self.structure().get_contents()
        
        if self._item is not None:
            self.item().found_by(self.target())


    
class Fight(Event):
    def __init__(self, source, target, min, max):
        Event.__init__(self, source, target)
        
        self._damage = random.randint(min, max)
    
    def damage(self):
        return self._damage
        
    def resistance(self):
        return self.source().resistance(self.damage())

    def description(self):
        nominative = self.target().gender().nominative()
        accusative = self.target().gender().accusative()
        possessive = self.target().gender().possessive_adjective()
        
        text = [('The %s hits you for %d damage.' % 
                 (str(self.target()).lower(),
                  self.damage() - self.resistance()))]
        
        if self.source().is_alive():
            text[0] += (' You kill %s.' % accusative)
            
            if self.target().gold() == 0:
                text += [('%s did not have any gold on %s.' % 
                          (nominative.capitalize(), accusative))]
            else:
                text += [('You loot %d gold from %s body.' %
                          (self.target().gold(), possessive))]
            
        return text
    
    def do(self):
        self.source().damage(self.damage())
        
        if self.source().is_alive():
            self.source().add_gold(self.target().gold())


    
class Blasting(Event):
    def __init__(self, source, target, chance):
        Event.__init__(self, source, target)
        
        self._keep = (random.randint(1, 100) <= chance)
    
    def keep(self):
        return self._keep
        
    def description(self):
        nominative = self.target().gender().nominative()
        accusative = self.target().gender().accusative()
        possessive = self.target().gender().possessive_adjective()
        
        text = ['You use a ruby to blast the %s.' % str(self.target()).lower()]
        
        if self.target().gold() == 0:
            text += [('%s did not have any gold on %s.' % 
                      (nominative.capitalize(), accusative))]
        else:
            text += [('You loot %d gold from %s body.' %
                      (self.target().gold(), possessive))]
        if not self.keep():
            text += ['The ruby vanishes!']
            
        return text
    
    def do(self):
        if not self.keep():
            self.source().consume_gem('fire')
            
        self.source().add_gold(self.target().gold())


    
class Hug(Event):
    def __init__(self, source, target):
        Event.__init__(self, source, target)
        
    def description(self):
        nominative = self.target().gender().nominative()
        
        text = '%s hugs you!' % nominative.capitalize()
        
        if self.target().gold() == 0:
            text += (' %s does not have any gold for you.' % 
                     (nominative.capitalize()))
        else:
            text += (' %s gives you %d gold.' %
                     (nominative.capitalize(), self.target().gold()))
            
        return [text]
    
    def do(self):
        self.source().add_gold(self.target().gold())


    
class Ashes(Event):
    def __init__(self, source, target):
        Event.__init__(self, source, target)
        
    def description(self):
        nominative = self.target().gender().nominative()
        
        if self.target().gold() == 0:
            text = ('%s did not have any gold.' % 
                    (nominative.capitalize()))
        else:
            text = ('You find %d gold in the ashes.' %
                    (self.target().gold()))
            
        return [text]
    
    def do(self):
        self.source().add_gold(self.target().gold())


    
class Bribe(Event):
    def __init__(self, source, target, gold):
        Event.__init__(self, source, target)
        
        self._gold = gold
    
    def gold(self):
        return self._gold

    def description(self):
        return ['The %s takes %d gold from you.' % 
                (str(self.target()).lower(), self.gold())]
            
    
    def do(self):
        self.source().lose_gold(self.gold())


    
class Run(Event):
    def __init__(self, source, target, chance, min, max):
        Event.__init__(self, source, target)
        
        self._success = (random.randint(0, 100) <= chance)
        self._damage  = random.randint(min, max)
    
    def damage(self):
        return self._damage

    def resistance(self):
        return self.source().resistance(self.damage())

    def success(self):
        return self._success
    
    def description(self):
        if self.success():
            text = ['You manage to get away.']
        else:
            text = [('The %s blocks you and hits you for %d damage.' %
                     (str(self.target()).lower(),
                      self.damage() - self.resistance()))]
            
        return text
    
    def do(self):
        if not self.success():
            self.source().damage(self.damage())



class Thirst(Event):
    def __init__(self, source, target, min, max):
        Event.__init__(self, source, target)
        
        self._damage = random.randint(min, max)
    
    def damage(self):
        return self._damage
        
    def resistance(self):
        return self.target().resistance(self.damage(), True)

    def description(self):
        return [('Thirst overcomes you. You suffer %d damage.' % 
                 (self.damage() - self.resistance()))]
    
    def do(self):
        self.target().damage(self.damage(), True)



class Tranquility(Event):
    def __init__(self, source, target, min, max):
        Event.__init__(self, source, target)
        
        self._heal = random.randint(min, max)
    
    def heal(self):
        return self._heal
        
    def description(self):
        return [('Life bristles around you. You rest for %d health.' % 
                 self.heal())]
    
    def do(self):
        self.target().heal(self.heal())
    
    

class Death(Event):
    def __init__(self, source, target, apocalyptic = False):
        Event.__init__(self, source, target)
    
        self._apocalyptic = apocalyptic
        
    def apocalyptic(self):
        return self._apocalyptic
    
    def description(self):
        if self.apocalyptic():
            return ['The apocalypse finally reached you.',
                    'Its magic flames burn through your body.',
                    'Not even ashes are left from it.',
                    'You die.']
        
        if not self.target().destroyed():
            return ['You die.']
        
        return []
    
    def do(self):
        self.target().set_health(0)
        
        if self.apocalyptic():
            self.target().destroyed = lambda: True
            
