'''Game model for Cursed Phoenix.'''

import math
import random

from concepts.application import Game
from concepts.map import Map
from concepts.units import Avatar
from concepts.geometry import Direction

import phoenix

from phoenix.score import HighScores
from phoenix.situations import DeadEnd, Outdoors, Battle
from phoenix.skills import *
from phoenix.structures import *
from phoenix.terrain import *



class PhoenixGame(Game):
    SKILLS = {'water': [Reflection, Mirage, Healing,
                        PoolOfLife, Cure, WaterShield],
              'earth': [TreeHugging, Survey, GemTrade,
                        Greed, Mining, StoneSkin],
              'fire' : [Blast, ScorchedEarth, Minting,
                        Climbing, FieryAura, Forging],
              'air'  : [Trajectory, Agility, Peace,
                        WindControl, Conjuring, Floating]}
    
    def __init__(self):
        Game.__init__(self)
        
        self._high = HighScores()

    def do_start(self):
        self.turn       = 0
        self._actions   = []
        self.map        = PhoenixMap(self, 49, 49)
        self.avatar     = PhoenixAvatar(self, 1000, self.map)
        self.apocalypse = Apocalypse(self, 15)
        
    def add_action(self, action):
        self._actions += [action]
    
    def fire_speed(self):
        return 100
    
    def set_highscores(self, high):
        self._high = high
    
    def highscores(self):
        return self._high
    
    def do_turn(self):
        self.turn += 1
        
        if self.turn > 100:
            self.apocalypse.do_turn(self.turn - 100)
            
        self.avatar.do_turn()
        
    def current_turn(self):
        return self.turn
    
    def current_map(self):
        return self.map
    
    def current_avatar(self):
        return self.avatar
    
    def current_situation(self):            
        return self.avatar.situation()
    
    def score(self):
        avatar = self.avatar
        stuff  = {}
        
        stuff['gold']     = avatar.gold()
        stuff['gems']     = {}
        stuff['feathers'] = {}
        
        for domain in ['water', 'earth', 'fire', 'air']:
            stuff['gems'][domain] = avatar.gems(domain)
            stuff['feathers'][domain] = avatar.feathers(domain)
        
        score = (stuff['gold'] + 
                 50 * reduce(lambda x, n: x + n, stuff['gems'].values(), 0))
        
        if self.has_won():
            score += 1000 * int(math.sqrt(reduce(lambda x, n: x * n,
                                                 stuff['feathers'].values(), 1)))
        
        self.highscores().register_score((score, stuff))
        
        return (score, stuff)
        
    def available_actions(self):
        if self.has_won():
            win = DirectAction('Ascend the Rainbow', None, [], [])
            win.set_situation(self.current_situation())
            
            return [win]
        
        if not self.avatar.is_alive():
            lose = DirectAction('Accept your Fate', None, [], [])
            lose.set_situation(self.current_situation())
            self.stop()
            
            return [lose]
            
        actions = self.current_situation().available_actions()
        
        if self.current_situation().allow_other_actions():
            for action in self._actions:
                action.set_situation(self.current_situation())
        
            actions += self._actions
        
        actions += [QuitAction(self.current_situation())]
        
        return actions
    
    def resolve_situation(self, situation):
        self.avatar.resolve_situation(situation)
    
    def skills(self, domain):
        return self.SKILLS[domain]

    def is_skill_selection(self):
        return isinstance(self.current_situation(), GainSkill)



class PhoenixAvatar(Avatar):
    def __init__(self, model, health, map):
        pos = map.start_tile()
        
        Avatar.__init__(self, model, pos.x, pos.y, map)

        self._health      = health
        self._gold        = 0
        self._gems        = {}
        self._feathers    = {}
        self._skills      = {}
        self._turn_hook   = []
        self._damage_hook = []
        self._inside      = False
    
    def forget_skills(self):
        self._skills      = {}
        self._turn_hook   = []
        self._damage_hook = []
        
        Outdoors.OTHER_ACTIONS = []
        Battle.OTHER_ACTIONS =  []
        phoenix.events.PhoenixFlight.see_direction = lambda event: False

        self.model().fire_speed = lambda: 100
        
        self.identify = lambda monster: False
        self.is_aquatic = lambda: False
        self.env_fire_resistance = lambda: 0
        self.is_forest_walking = lambda: False
        self.forest_riches = lambda: 0
        self.survey = lambda: 0
        self.can_mine = lambda: False
        self.mining   = lambda: 0
        self.minting = lambda: 0
        self.is_climbing = lambda: False
        self.fiery_aura = lambda: 0
        self.forging = lambda: 0
        self.agility = lambda: 0
        self.is_peace = lambda: False
        self.conjuring = lambda: False
        self.is_floating = lambda: False
        self.is_evading_traps = lambda: False
        self.mountain_riches = lambda: False
        
    def mountain_riches(self):
        return False

    def destroyed(self):
        return False
    
    def identify(self, monster):
        return False
    
    def forging(self):
        return 0
    
    def minting(self):
        return 0
    
    def conjuring(self):
        return False
    
    def agility(self):
        return 0
    
    def can_mine(self):
        return False
    
    def mining(self):
        return 0
    
    def fiery_aura(self):
        return 0
    
    def is_climbing(self):
        return False
    
    def is_floating(self):
        return False
    
    def is_evading_traps(self):
        return False
    
    def is_peace(self):
        return False
    
    def is_aquatic(self):
        return False
    
    def is_forest_walking(self):
        return False
    
    def survey(self):
        return 1.0
    
    def forest_riches(self):
        return 0
    
    def env_fire_resistance(self):
        return 0
    
    def is_inside(self):
        return self._inside
    
    def go_inside(self):
        self._inside = True
    
    def go_outside(self):
        self._inside = False
    
    def do_turn(self):
        for function in self._turn_hook:
            function(self.model())
        
        self.do_events()
    
    def do_every_turn(self, function):
        self._turn_hook.append(function)
    
    def do_when_resisting(self, function):
        self._damage_hook.append(function)
        
    def add_gem(self, domain):
        self._gems[domain] = self.gems(domain) + 1
        
    def add_gems(self, domain, num):
        if domain == 'all':
            for domain in ['water', 'earth', 'fire', 'air']:
                self.add_gems(domain, num)
        else:
            self._gems[domain] = self.gems(domain) + num
        
    def consume_gem(self, domain):
        self._gems[domain] = self.gems(domain) - 1
        
    def gems(self, domain):
        return self._gems.get(domain, 0)
        
    def add_skill(self, skill):
        domain = skill.domain()
        self._skills[domain] = self.skills(domain) + [skill]
        skill.aquired_by(self)
        
    def skills(self, domain = None):
        if domain is None:
            return reduce(lambda a, b: a + b, self._skills.values(), [])
        
        return self._skills.get(domain, [])
    
    def add_feather(self, domain):
        self._feathers[domain] = self.feathers(domain) + 1
        
    def feathers(self, domain):
        return self._feathers.get(domain, 0)
    
    def is_alive(self):
        return (self.health() > 0)
    
    def health(self):
        return self._health
    
    def gold(self):
        return self._gold
    
    def set_health(self, health):
        self._health = health
    
    def scorch(self, degree):
        for tile in self.model().current_map().tiles_in_radius(self, 1):
            tile.terrain.scorch(degree)
            
    def situation(self):
        if self.is_alive():
            return self.current_tile().situation()
        else:
            return DeadEnd(self.model())

    def resolve_situation(self, situation):
        if self.is_alive():
            return self.current_tile().resolve_situation(situation)
        
    def collect_self_events(self):
        if self.is_alive():
            return []
        else:
            return [Death(self, self)]
    
    def heal(self, num):
        self._health += num

    def resistance(self, num, env_fire = False):
        total = 0
        
        if env_fire:
            total += num * self.env_fire_resistance() / 100
            
        for function in self._damage_hook:
            total += function(num)
            
        if total >= num:
            return num - 1
        
        return total
        
    def damage(self, num, env_fire = False):
        num -= self.resistance(num, env_fire)
        self._health -= num
    
    def hurt(self, num):
        self._health -= num
        
    def add_gold(self, num):
        self._gold += num
    
    def lose_gold(self, num):
        self._gold -= num
        
    def current_hp(self):
        return self.hp
    
    def __str__(self):
        if self.is_alive():
            return 'Avatar'
        elif self.destroyed():
            return 'Destroyed Avatar'
        else:
            return 'Dead Avatar'


    
class PhoenixMap(Map):
    def closest_tile(self, condition, pos):
        closest = lambda tile: -tile.distance(pos)
        tile    = self.top_tile(closest, condition)
        
        return tile
                
    def reveal_secrets(self, pos, radius):
        for tile in self.tiles_in_radius(pos, radius):
            if tile.is_built():
                tile.structure().unhide()
    
    def create_lake(self, pos, radius, max_depth):
        for tile in self.tiles_in_radius(pos, radius, True):
            if tile.distance(pos, True) > radius * 1.2:
                continue

            if (tile.distance(pos, True) > radius * 0.8 and 
                random.randint(0, 3) <= 1):
                continue
            
            factor = 1 - (tile.distance(pos, True) / radius
                          - random.randint(-10, 40) / 100.0)
            
            tile.become(Lake, int(factor * max_depth))
    
    def create_forest(self, pos, radius):
        for tile in self.tiles_in_radius(pos, radius, True):
            if tile.distance(pos, True) > radius * 1.2:
                continue

            if (tile.distance(pos, True) > radius * 0.8 and 
                random.randint(0, 3) <= 1):
                continue
            
            factor = 1 - (tile.distance(pos, True) / radius
                          - random.randint(-10, 40) / 100.0)
            
            tile.become(Forest, factor > random.random())
    
    def create_hills(self, pos, radius):
        for tile in self.tiles_in_radius(pos, radius, True):
            if tile.distance(pos, True) > radius * 1.2:
                continue

            if (tile.distance(pos, True) > radius * 0.8 and 
                random.randint(0, 3) <= 1):
                continue
            
            factor = 1 - (tile.distance(pos, True) / radius
                          - random.randint(-30, 40) / 100.0)
            
            if factor > random.random():
                tile.become(Hills)
    
    def tile_between(self, max, min, distance = 0,
                     unbuilt = False, land = False, water = False, kind = None):
        accept = False
        
        while not accept:
            x      = random.randint(distance, self.width - distance - 1)
            y      = random.randint(min, max)
            tile   = self.tile(Position(x, y))
            accept = True
            
            if kind is not None:
                accept = isinstance(tile.terrain, kind)
                continue
            
            if unbuilt and tile.is_built():
                accept = False
                
            if land and (tile.water_depth() > 0 or tile.terrain.special_land()):
                accept = False
                
            if water and tile.water_depth() == 0:
                accept = False
                
        return tile
    
    def free_tile_in(self, domain, kind = None):
        water = self.height - 1
        earth = self.height * 3 / 4
        fire  = self.height * 1 / 2
        air   = self.height * 1 / 4
        top   = 0
        
        if domain == 'all':
            return self.tile_between(water - 3, top + 3, 0, True, True, False, kind)
        elif domain == 'water':
            return self.tile_between(water - 3, earth, 0, True, True, False, kind)
        elif domain == 'earth':
            return self.tile_between(earth - 3, fire, 0, True, True, False, kind)
        elif domain == 'fire':
            return self.tile_between(fire - 3, air, 0, True, True, False, kind)
        elif domain == 'air':
            return self.tile_between(air - 3, top + 3, 0, True, True, False, kind)
        
        return None
    
    def distribute(self, domain, num, structure, kind = None):
        for nr in range(num):
            tile = self.free_tile_in(domain, kind)
            tile.build(structure)
    
    def create_stream(self, start_y, end_y, min_y, max_y, fuzzy, kind):
        dy = 0
        
        for x in range(self.width):
            if random.randint(1, 100) <= fuzzy:
                dy += random.randint(-1, 1)
            
            cy = start_y + (end_y - start_y) * x / self.width + dy
            h  = random.randint(min_y, max_y)
            
            for y in range(cy, cy + h + 1):
                self.tile(Position(x, y)).become(Lava)
            
    def generate(self):
        water = self.height - 1
        earth = self.height * 3 / 4
        fire  = self.height * 1 / 2
        air   = self.height * 1 / 4
        top   = 0

        for tile in self:
            if (random.randint(-self.height, self.height / 4) > 
                abs(tile.y - self.height * 3 / 4)):
                tile.become(Forest)
            elif (random.randint(-self.height, self.height / 4) > 
                abs(tile.y - self.height * 1 / 2)):
                tile.become(Hills)
            elif (random.randint(-self.height, self.height / 4) > 
                  tile.y * 4 - abs(tile.x - self.width / 2) * 4 + 
                  self.width / 2 + self.height / 8):
                tile.become(Mountain)
                if random.randint(0, 1):
                    tile.become(Mountain)
                else:
                    tile.become(Glacier)
            elif (random.randint(0, self.height * 1 / 4) < air - tile.y):
                if random.randint(0, 1):
                    tile.become(Mountain)
                else:
                    tile.become(Glacier)
            else:
                tile.become(Plains, tile.y * 27 / self.height + 
                            random.randint(-15, 10))
        
        for nr in range(5):
            hills_tile = self.tile_between(fire + 4, fire - 4, 0, False, True)
            self.create_hills(hills_tile, 2)
            
            hills_tile = self.tile_between(fire, air + 4, 0, False, True)
            self.create_hills(hills_tile, 3)
            
            hills_tile = self.tile_between(fire, air, 0, False, True)
            self.create_hills(hills_tile, 2)
            
        hills_tile = self.tile_between(fire + 4, fire - 4, 0, False, True)
        self.create_hills(hills_tile, 4)
        hills_tile.build(DragonCave)
        
        forest_tile = self.tile_between(earth - 3, earth - 8, 0, False, True)
        self.create_forest(forest_tile, 3)
        
        forest_tile = self.tile_between(earth - 3, earth - 8, 0, False, True)
        self.create_forest(forest_tile, 3)
        forest_tile.build(FountainOfYouth)
        
        forest_tile = self.tile_between(earth - 3, earth - 8, self.width / 5, False, True)
        self.create_forest(forest_tile, 3)
        
        forest_tile = self.tile_between(earth - 3, earth - 8, self.width / 3, False, True)
        self.create_forest(forest_tile, 6)
        forest_tile.build(TreeOfAges)
        
        lake_tile = self.tile_between(water - 5, earth + 3, 5)
        self.create_lake(lake_tile, 3, 100)
        
        lake_tile = self.tile_between(earth - 3, earth - 8, 5, False, True)
        self.create_lake(lake_tile, 2, 50)
        
        lake_tile = self.tile_between(water - 10, earth + 2, 5, False, True)
        self.create_lake(lake_tile, 2, 90)
        
        self.create_stream(air + 1, air + 4, 2, 3, 10, Lava)
        
        crystal_tile = self.tile_between(air - 1, air - 2, self.width / 4)
        crystal_tile.build(CrystalPlatform)
        
        for tile in self.tiles_in_radius(crystal_tile, 1):
            tile.become(Plains, 5)
            
        self.create_forest(forest_tile, 3)
        
        tile = self.top_tile(lambda tile: tile.water_depth())
        tile.build(SunkenCity)
        
        self._start_tile = self.tile_between(water, water - 2, self.width / 4,
                                             True, True)
        self._start_tile.build(Hut)

        town_tile = self.tile_between(water - 4, water - 5, self.width * 2 / 5,
                                      True, True)
        town_tile.build(Town)
        
        self.tile(Position(self.width / 2, 0)).become(RainbowMountain)
        
        self.distribute('water', 1, WishingWell)
        
        self.distribute('water', 3, BlackMarket)
        self.distribute('earth', 4, BlackMarket)
        self.distribute('fire',  3, BlackMarket)
        self.distribute('air',   2, BlackMarket)
        
        self.distribute('water', 12, AlchemistsHut)
        self.distribute('earth',  8, AlchemistsHut)
        self.distribute('fire',   2, AlchemistsHut)
        self.distribute('air',    4, AlchemistsHut)

        self.distribute('water', 1, Town)
        self.distribute('earth', 1, Town)
        self.distribute('fire',  1, Town)
        self.distribute('air',   1, Town)

        self.distribute('all',  15, Mine, Hills)
        
        self.distribute('fire', 1, LavaForge,  Lava)
        self.distribute('air',  1, IvoryTower, Glacier)
        
        for domain in (['water'] * 3) + ['earth', 'air'] * 2:
            tile = self.tile_between(water - 3, earth, 0, True, True)
            tile.build(PhoenixNest, domain)
        
        for domain in (['earth'] * 3) + ['water', 'fire'] * 2:
            tile = self.tile_between(earth - 3, fire, 0, True, True)
            tile.build(PhoenixNest, domain)
        
        for domain in (['fire'] * 3) + ['earth', 'air'] * 2:
            tile = self.tile_between(fire - 3, air, 0, True, True)
            tile.build(PhoenixNest, domain)
        
        for domain in (['air'] * 3) + ['water', 'fire'] * 2:
            tile = self.tile_between(air - 3, top + 3, 0, True, True)
            tile.build(PhoenixNest, domain)



class Apocalypse:
    def __init__(self, model, speed):
        self._model = model
        self._speed = speed
    
    def model(self):
        return self._model
    
    def speed(self):
        return self._speed * self.model().fire_speed() / 100
    
    def map(self):
        return self.model().current_map()
    
    def do_turn(self, turn):
        height = self.map().height - (turn / self.speed()) - 1
        
        for row in range(height, self.map().height):
            for tile in self.map().reverse[row]:
                if random.randint(-self.speed() / 2, row - height) > 0:
                    tile.terrain.scorch(1, True)
                    
                    if tile.livability() < 50:
                        tile.destroy_buildings()

                    if tile.livability() <= 0:
                        tile.become(Apocalyptic)
                        