'''Items in the Phoenix game.'''

import random

import phoenix.situations

def GainSkill(*foo):
    return phoenix.situations.GainSkill(*foo)



class Item:
    def __init__(self, model, name):
        self._name  = name
        self._model = model
    
    def number(self):
        return 1
        
    def description(self):
        if self.number() > 1:
            return '%d %s' % (self.number(), self.name())
        
        if self.name()[0] in 'aeiou':
            return 'an %s' % self.name()
        else:
            return 'a %s' % self.name()
        
    def name(self):
        return self._name
        
    def model(self):
        return self._model
    
    def found_by(self, avatar):
        pass
    
    def find_description(self):
        return []
        
    def postfix(self):
        return None
    
    def __str__(self):
        if self.postfix() is not None:
            return '%s %s' % (self.description(), self.postfix())
        
        return self.description()



class DomainItem(Item):
    def __init__(self, model, domain, name):
        Item.__init__(self, model, name)
        
        self._domain = domain
        
    def domain(self):
        return self._domain
    
    
    
class Gem(DomainItem):
    NAMES = {'water': 'sapphire',
             'earth': 'emerald',
             'fire' : 'ruby',
             'air'  : 'topaz'}
    
    def __init__(self, model, domain):
        DomainItem.__init__(self, model, domain, self.NAMES[domain])
            
    def found_by(self, avatar):
        avatar.add_gem(self.domain())
    
    def find_description(self):
        return [('Your power in the domain of %s increases!' % 
                 self.domain())]



class Gems(DomainItem):
    NAMES = {'water': 'sapphires',
             'earth': 'emeralds',
             'fire' : 'rubies',
             'air'  : 'topazes'}

    def __init__(self, model, domain, num):
        DomainItem.__init__(self, model, domain, self.NAMES[domain])
        
        self._number = num
    
    def number(self):
        return self._number
    
    def found_by(self, avatar):
        for nr in range(self.number()):
            avatar.add_gem(self.domain())
    
    def find_description(self):
        return [('Your power in the domain of %s increases!' % 
                 self.domain())]
        
    
    
class Feather(DomainItem):
    DOMAIN_MAXED = {'water': 'nearly drown in it',
                    'earth': 'turn to stone for a moment',
                    'fire' : 'feel a burning sensation',
                    'air'  : 'get blown upwards, then float down'}
    
    def __init__(self, model, domain):
        DomainItem.__init__(self, model, domain, 'feather')
    
        self._situation = None
        
    def situation(self):
        return self._situation
    
    def found_by(self, avatar):
        self._situation = GainSkill(self.model(), avatar, self.domain())
        
        avatar.add_feather(self.domain())
        
        if self.situation().can_increase_skill():
            avatar.current_tile().set_situation(self.situation())
        
    def find_description(self):
        power = 'You take the feather. Its power flows into you!'
            
        if self.situation().is_completly_maxed():
            return [power, 'It nearly tears you apart!']
        
        if self.situation().is_domain_maxed():
            return [power, 'You %s!' % self.DOMAIN_MAXED[self.domain()]]
        
        return [power, 'You gain access to its magic!']
    
    
    
class PreciousStone(Item):
    NAMES = ['opal', 'amethyst', 'pearl', 'turquise', 'obsidian']
    
    def __init__(self, model, worth):
        Item.__init__(self, model, random.choice(self.NAMES))
        
        self._worth = worth
        
    def worth(self):
        return self._worth
    
    def found_by(self, avatar):
        avatar.add_gold(self.worth())
    
    def postfix(self):
        return ('worth %d gold' % self.worth())




class Gold(Item):
    def __init__(self, model, worth):
        Item.__init__(self, model, '%d gold' % worth)
        
        self._worth = worth
        
    def worth(self):
        return self._worth
    
    def found_by(self, avatar):
        avatar.add_gold(self.worth())


    
class BlackMarketTrade:
    def __init__(self, avatar, market, sale, num, domain, gold):
        self._avatar = avatar
        self._market = market
        self._sale   = sale
        self._gold   = gold
        self._num    = num
        self._domain = domain
    
    def avatar(self):
        return self._avatar
    
    def market(self):
        return self._market
    
    def is_sale(self):
        return self._sale
    
    def gold(self):
        return self._gold
    
    def num(self):
        return self._num
    
    def domain(self):
        return self._domain
    
    def __str__(self):
        if self.is_sale():
            buy = 'S:'
        else:
            buy = 'B:'
        
        if self.num() == 1:
            item = Gem(None, self.domain())
        else:
            item = Gems(None, self.domain(), self.num())
        
        return '%s %s' % (buy, str(item))
    
