'''Situations in the Phoenix game.'''

from concepts.actions import *
from concepts.application import Situation
from concepts.geometry import COMPASS

from phoenix.actions import *
from phoenix.events import *



class Battle(Situation):
    OTHER_ACTIONS = []

    def __init__(self, model, avatar, monster):
        Situation.__init__(self, model)
        
        self._avatar  = avatar
        self._monster = monster
            
    def allow_other_actions(self):
        return False
    
    def avatar(self):
        return self._avatar
    
    def monster(self):
        return self._monster
    
    def combat_info(self):
        max  = self.monster().strength()
        min  = max * 3 / 4
        max -= self.avatar().resistance(max)
        min -= self.avatar().resistance(min)

        if self.avatar().identify(self.monster()):
            return '%dg [%d-%d]' % (self.monster().gold(), min, max)
        
        return ''
    
    def run_info(self):
        damage = self.monster().strength()
        max    = damage * 4 / 8
        min    = damage * 1 / 4
        max   -= self.avatar().resistance(max)
        min   -= self.avatar().resistance(min)
        chance = 100 - self.monster().agility() + self.avatar().agility()
        
        if chance > 100:
            chance = 100

        if self.avatar().identify(self.monster()):
            return '%d%% [%d-%d]' % (chance, min, max)
        
        return ''
    
    def bribe_info(self):
        if self.avatar().identify(self.monster()):
            return '%dg' % self.monster().greed()
        
        return ''
        
    def fight(self):
        damage = self.monster().strength()
        
        self.model().add_events([Fight(self.avatar(), self.monster(),
                                       damage * 3 / 4, damage)])
        self.resolve()
        
    def blast(self, chance):
        self.model().add_events([Blasting(self.avatar(), self.monster(), 
                                          chance)])
        self.resolve()
    
    def run(self):
        damage = self.monster().strength()
        chance = 100 - self.monster().agility() + self.avatar().agility()
        
        if chance > 100:
            chance = 100
            
        run = Run(self.avatar(), self.monster(),
                  chance, damage * 1 / 4, damage * 4 / 5)
        
        self.model().add_events([run])
        
        if run.success():
            self.resolve()
    
    def bribe(self):
        self.model().add_events([Bribe(self.avatar(), self.monster(),
                                       self.monster().greed())])
        
        self.resolve()
        
    def available_actions(self):                
        for action in self.OTHER_ACTIONS:
            action.set_situation(self)

        return ([FightAction(self), RunAction(self), BribeAction(self)] +
                self.OTHER_ACTIONS)



class Outdoors(Situation):
    OTHER_ACTIONS = []
    
    def available_actions(self):
        actions = []
        
        for dir in COMPASS.values():
            actions.append(MoveAction(self, dir))
        
        for action in self.OTHER_ACTIONS:
            action.set_situation(self)
            
        actions += [NullAction(self)]
        actions += self.OTHER_ACTIONS
        
        return actions



class BlackMarketSituation(Situation):
    def __init__(self, model, avatar, market):
        Situation.__init__(self, model)
        
        self._avatar = avatar
        self._market = market
        
    def avatar(self):
        return self._avatar
    
    def market(self):
        return self._market
    
    def available_actions(self):
        actions = []
        
        for dir in COMPASS.values():
            actions.append(MoveAction(self, dir))

        for trade in self.market().trades():
            actions += [BlackMarketAction(self, trade)]
        
        return actions



class AlchemistSituation(Situation):
    def __init__(self, model, avatar, market):
        Situation.__init__(self, model)
        
        self._avatar = avatar
        self._market = market
        
    def avatar(self):
        return self._avatar
    
    def market(self):
        return self._market
    
    def available_actions(self):
        actions = []
        
        for dir in COMPASS.values():
            actions.append(MoveAction(self, dir))
        
        for potion in self.market().potions():
            actions += [AlchemistAction(self, potion, self.market())]
        
        return actions



class TownSituation(Situation):
    def __init__(self, model, avatar, market):
        Situation.__init__(self, model)
        
        self._avatar = avatar
        self._market = market
        
    def avatar(self):
        return self._avatar
    
    def market(self):
        return self._market
    
    def available_actions(self):
        actions = []
        
        for dir in COMPASS.values():
            actions.append(MoveAction(self, dir))
        
        for item in self.market().maps():
            actions += [TownAction(self, item, self.market())]
        
        return actions



class StructureSituation(Situation):
    def __init__(self, structure):
        Situation.__init__(self, structure.model())
        
        self._structure = structure
        
    def structure(self):
        return self._structure
    
    def available_actions(self):
        actions = []
        
        for dir in COMPASS.values():
            actions.append(MoveAction(self, dir))
        
        for params in self.structure().actions():
            action = DirectAction(*params)
            action.set_situation(self)
            actions.append(action)
        
        return actions



class DirectSituation(Situation):
    def __init__(self, model, params, move = True):
        Situation.__init__(self, model)
        
        self._params = params
        self._move   = move
        
    def params(self):
        return self._params
    
    def move(self):
        return self._move
    
    def available_actions(self):
        actions = []
        
        if self.move():
            for dir in COMPASS.values():
                actions.append(MoveAction(self, dir))
        
        for params in self.params():
            params[3].append(lambda: self.resolve())
            action = DirectAction(*params)
            action.set_situation(self)
            actions.append(action)
        
        return actions



class GainSkill(Situation):
    def __init__(self, model, avatar, domain):
        Situation.__init__(self, model)
        
        self._avatar = avatar
        self._domain = domain
    
    def allow_other_actions(self):
        return False

    def avatar(self):
        return self._avatar
    
    def domain(self):
        return self._domain
    
    def is_domain_maxed(self):
        return len(self.avatar().skills(self.domain())) >= 3
    
    def is_completly_maxed(self):
        return len(self.avatar().skills()) >= 7
    
    def can_increase_skill(self):
        return (not self.is_domain_maxed() and not self.is_completly_maxed())
    
    def available_actions(self):
        actions  = [GainSkillAction(self, skill)
                    for skill in self.model().skills(self.domain())]        
        actions += [RefuseSkillAction(self)]
        
        return actions



class DeadEnd(Situation):
    def available_actions(self):
        return []
    
    def allow_other_actions(self):
        return False
