'''Skills in the Phoenix game.'''

import math
import random

import phoenix.actions
import phoenix.situations
import phoenix.events



class Skill:
    def __init__(self, name, domain, require):
        self._name    = name
        self._domain  = domain
        self._require = require
        self._avatar  = None
        
    def name(self):
        return self._name
    
    def domain(self):
        return self._domain
    
    def avatar(self):
        return self._avatar
    
    def power(self):
        if self.avatar() is None:
            return 0
        
        power = self.avatar().gems(self.domain())
        
        if self.power_cap() is not None and power > self.power_cap():
            return self.power_cap()
        
        return power
    
    def power_cap(self):
        return None
    
    def aquired_by(self, avatar):
        self._avatar = avatar
        self.start_magic(avatar)
    
    def required_feathers(self):
        return self._require
    
    def effect(self):
        raise 'Not implemented'
    
    def start_magic(self, avatar):
        raise 'Not implemented'
    
    def info(self):
        raise 'Not implemented'
    
    def description(self):
        raise 'Not implemented'
        
    def __str__(self):
        return self._name
    
    def __eq__(self, other):
        return str(self) == str(other)



class Mirage(Skill):
    def __init__(self):
        Skill.__init__(self, 'Mirage', 'water', 1)
    
    def effect(self):
        return (self.power() + 25) / 30
    
    def info(self):
        return 'Rad %d' % (self.effect())

    def description(self):
        return 'Reveals hidden things'
    
    def start_magic(self, avatar):
        avatar.do_every_turn(lambda model:
                             model.current_map().reveal_secrets(avatar,
                                                                self.effect()))



class Reflection(Skill):
    def __init__(self):
        Skill.__init__(self, 'Reflection', 'water', 1)
    
    def effect(self):
        return int(math.sqrt(self.power() + 1) * 20)
    
    def info(self):
        return 'Str %d' % (self.effect())

    def description(self):
        return 'See enemies clearly'
    
    def start_magic(self, avatar):
        avatar.identify = lambda monster: monster.strength() <= self.effect()



class Healing(Skill):
    def __init__(self):
        Skill.__init__(self, 'Healing', 'water', 2)
    
    def effect(self):
        return 50 + int(math.log(self.power() + 1, 1.25) * 4)
    
    def info(self):
        return '+%d' % (self.effect())

    def description(self):
        return 'Heal yourself a bit'
    
    def start_magic(self, avatar):
        phoenix.situations.Outdoors.OTHER_ACTIONS += [phoenix.actions.HealAction(None, lambda: self.effect(),
                                                                                 lambda: self.effect())]



class PoolOfLife(Skill):
    def __init__(self):
        Skill.__init__(self, 'Pool of Life', 'water', 2)
    
    def effect(self):
        return int(math.log(self.power() + 1, 2) * 3) + 1
    
    def info(self):
        return '+0-%d' % self.effect()

    def description(self):
        return 'Lakes heal you'
    
    def start_magic(self, avatar):
        avatar.is_aquatic = lambda: True
        avatar.do_every_turn(lambda model: 
                             avatar.heal(self.effect() *
                                         avatar.current_tile().water_depth() /
                                         100))



class Cure(Skill):
    def __init__(self):
        Skill.__init__(self, 'Cure', 'water', 4)
    
    def effect(self):
        return int(math.log(self.power() + 1, 1.75) * 2)
    
    def info(self):
        return '+%d/t' % (self.effect())

    def description(self):
        return 'Heals you every turn'
    
    def start_magic(self, avatar):
        avatar.do_every_turn(lambda model: avatar.heal(self.effect()))



class WaterShield(Skill):
    def __init__(self):
        Skill.__init__(self, 'Water Shield', 'water', 4)
    
    def power_cap(self):
        return 50
    
    def effect(self):
        return 25 + self.power()

    def info(self):
        return '%d%%' % (self.effect())

    def description(self):
        return 'Survive in hot places'
    
    def start_magic(self, avatar):
        avatar.env_fire_resistance = lambda: self.effect()



class TreeHugging(Skill):
    def __init__(self):
        Skill.__init__(self, 'Tree Hugging', 'earth', 1)
    
    def effect(self):
        return 5 + int(math.sqrt(self.power()) * 2.5)
    
    def info(self):
        return '%d%%' % (self.effect())

    def description(self):
        return 'Enjoy forest riches'
    
    def start_magic(self, avatar):
        avatar.is_forest_walking = lambda: True
        avatar.forest_riches = lambda: self.effect()



class Survey(Skill):
    def __init__(self):
        Skill.__init__(self, 'Survey', 'earth', 1)
    
    def effect(self):
        return 1 + math.sqrt(self.power()) / 4
    
    def info(self):
        return '%.2fx' % (self.effect())

    def description(self):
        return 'More deposits per tile'
    
    def start_magic(self, avatar):
        avatar.survey = lambda: self.effect()



class GemTrade(Skill):
    def __init__(self):
        Skill.__init__(self, 'Gem Trade', 'earth', 2)
    
    def effect(self):
        return int(math.sqrt(self.power()) * 3)
    
    def info(self):
        return '%d%%' % (self.effect())

    def description(self):
        return 'Emeralds buy other gems'
    
    def start_magic(self, avatar):
        phoenix.situations.Outdoors.OTHER_ACTIONS += [phoenix.actions.TradeGemAction(None, self.domain(),
                                                      lambda: self.effect())]



class Greed(Skill):
    def __init__(self):
        Skill.__init__(self, 'Greed', 'earth', 2)
        
        self._avatar = None
        
    def avatar(self):
        return self._avatar
    
    def effect(self):
        return int(self.avatar().gold() * math.sqrt(self.power()) / 10000)
    
    def info(self):
        return '+%d/t' % (self.effect())

    def description(self):
        return 'Charge interest on gold'
    
    def start_magic(self, avatar):
        self._avatar = avatar

        avatar.do_every_turn(lambda model: avatar.add_gold(self.effect()))



class StoneSkin(Skill):
    def __init__(self):
        Skill.__init__(self, 'Stone Skin', 'earth', 4)
    
    def power_cap(self):
        return 45
    
    def effect(self):
        return 5 + self.power()
    
    def info(self):
        return '%d%%' % (self.effect())

    def description(self):
        return 'Makes you resist damage'
    
    def start_magic(self, avatar):
        avatar.do_when_resisting(lambda damage: damage * self.effect() / 100)



class Mining(Skill):
    def __init__(self):
        Skill.__init__(self, 'Mining', 'earth', 4)
    
    def power_cap(self):
        return 100
    
    def effect(self):
        return self.power() / 2
    
    def info(self):
        return '%d%%' % (self.effect())

    def description(self):
        return 'Mine for treasure'
    
    def start_magic(self, avatar):
        avatar.can_mine = lambda: True
        avatar.mining   = lambda: self.effect()



class Blast(Skill):
    def __init__(self):
        Skill.__init__(self, 'Blast', 'fire', 1)
    
    def power_cap(self):
        return 40
    
    def effect(self):
        return 75 + self.power() / 2
    
    def info(self):
        return '%d%%' % (self.effect())

    def description(self):
        return 'Blast your enemies'
    
    def start_magic(self, avatar):
        phoenix.situations.Battle.OTHER_ACTIONS += [phoenix.actions.BlastAction(None, lambda: self.effect())]

 
 
class ScorchedEarth(Skill):
    def __init__(self):
        Skill.__init__(self, 'Scorched Earth', 'fire', 1)
    
    def effect(self):
        return int(math.log(self.power() + 1, 2) / 5) + 1
    
    def info(self):
        return '%dx' % (self.effect())

    def description(self):
        return 'Burns down pesky trees'
    
    def start_magic(self, avatar):
        phoenix.situations.Outdoors.OTHER_ACTIONS += [phoenix.actions.ScorchAction(None, lambda: self.effect())]

 
 
class Minting(Skill):
    def __init__(self):
        Skill.__init__(self, 'Minting', 'fire', 2)
    
    def power_cap(self):
        return 90
    
    def effect(self):
        return self.power() / 3 * 25
    
    def info(self):
        return '+%d' % (self.effect())

    def description(self):
        return 'Mint gold in lava'
    
    def start_magic(self, avatar):
        avatar.minting = lambda: self.effect()



class Climbing(Skill):
    def __init__(self):
        Skill.__init__(self, 'Climbing', 'fire', 2)
    
    def power_cap(self):
        return 50
    
    def effect(self):
        return self.power() / 5
    
    def info(self):
        return '%d%%' % (self.effect())

    def description(self):
        return 'Move through mountains'
    
    def start_magic(self, avatar):
        avatar.is_climbing = lambda: True
        avatar.mountain_riches = lambda: random.randint(1, 100) <= self.effect()

 
 
class FieryAura(Skill):
    def __init__(self):
        Skill.__init__(self, 'Fiery Aura', 'fire', 4)
    
    def power_cap(self):
        return 75
    
    def effect(self):
        return self.power() * 2
    
    def info(self):
        return 'Str %d' % (self.effect())

    def description(self):
        return 'Kills weaker enemies'
    
    def start_magic(self, avatar):
        avatar.fiery_aura = lambda: self.effect()
 
  
  
class Forging(Skill):
    def __init__(self):
        Skill.__init__(self, 'Forging', 'fire', 4)
    
    def power_cap(self):
        return 100
    
    def effect(self):
        return int(math.sqrt(self.power()) / 3)
    
    def info(self):
        return '+%d' % (self.effect())

    def description(self):
        return 'Forge rubies in lava'
    
    def start_magic(self, avatar):
        avatar.forging = lambda: self.effect()



class Trajectory(Skill):
    def __init__(self):
        Skill.__init__(self, 'Trajectory', 'air', 1)
        
    def power_cap(self):
        return 40
    
    def effect(self):
        return 10 + self.power() * 2
    
    def info(self):
        return '%d%%' % (self.effect())
    
    def description(self):
        return 'See where birds fly'
    
    def start_magic(self, avatar):
        phoenix.events.PhoenixFlight.see_direction = (lambda event:
                                                      random.randint(1, 100) <= self.effect())



class Agility(Skill):
    def __init__(self):
        Skill.__init__(self, 'Agility', 'air', 1)
    
    def power_cap(self):
        return 45
    
    def effect(self):
        return 5 + self.power()
    
    def info(self):
        return '+%d%%' % (self.effect())

    def description(self):
        return 'Run faster than most'
    
    def start_magic(self, avatar):
        avatar.agility = lambda: self.effect()

 
 
class Peace(Skill):
    def __init__(self):
        Skill.__init__(self, 'Peace', 'air', 2)
    
    def power_cap(self):
        return 100
    
    def effect(self):
        return self.power()
    
    def info(self):
        return '%d%%' % (self.effect())

    def description(self):
        return 'Create an aura of peace'
    
    def start_magic(self, avatar):
        avatar.is_peace = lambda: random.randint(1, 100) <= self.effect()



class WindControl(Skill):
    def __init__(self):
        Skill.__init__(self, 'Wind Control', 'air', 2)
    
    def power_cap(self):
        return 100
    
    def effect(self):
        return -self.power() * 3 / 5
    
    def info(self):
        return '%d%%' % (self.effect())

    def description(self):
        return 'Slow down the apocalypse'
    
    def start_magic(self, avatar):
        avatar.model().fire_speed = lambda: 100 - self.effect()



class Conjuring(Skill):
    def __init__(self):
        Skill.__init__(self, 'Conjuring', 'air', 4)
    
    def power_cap(self):
        return 100
    
    def effect(self):
        return math.sqrt(self.power())
    
    def info(self):
        return '%d%%' % (self.effect())

    def description(self):
        return 'Gems from thin air'
    
    def start_magic(self, avatar):
        avatar.conjuring = lambda: random.randint(1, 100) <= self.effect()



class Floating(Skill):
    def __init__(self):
        Skill.__init__(self, 'Floating', 'air', 4)
    
    def power_cap(self):
        return 50

    def effect(self):
        return self.power() * 2
    
    def info(self):
        return '%d%%' % (self.effect())

    def description(self):
        return 'No terrain can stop you'
    
    def start_magic(self, avatar):
        avatar.is_floating      = lambda: True
        avatar.is_evading_traps = lambda: random.randint(1, 100) <= self.effect()
    