'''Structures (buildings) in the Phoenix game.'''

import math
import random

from concepts.map import Structure

from phoenix.items import *
from phoenix.skills import *

import phoenix.situations
import phoenix.events
  
def BlackMarketSituation(*foo):
    return phoenix.situations.BlackMarketSituation(*foo)
  
def TownSituation(*foo):
    return phoenix.situations.TownSituation(*foo)
  
def AlchemistSituation(*foo):
    return phoenix.situations.AlchemistSituation(*foo)
  
def StructureSituation(*foo):
    return phoenix.situations.StructureSituation(*foo)
  
def Outdoors(*foo):
    return phoenix.situations.Outdoors(*foo)
  
def FindStructure(*foo):
    return phoenix.events.FindStructure(*foo)
  
def SeeStructure(*foo):
    return phoenix.events.SeeStructure(*foo)



class PhoenixStructure(Structure):
    def __init__(self, model, tile, domain, hidden, outside, item, name):
        Structure.__init__(self, model, tile, hidden, outside, name)
        
        self._domain    = domain
        self._item      = item
        self._situation = None
        
    def domain(self):
        return self._domain
    
    def is_dark_place(self):
        return False
    
    def is_wooden(self):
        return False
    
    def get_contents(self):
        item = self._item
        self._item = None
        
        return item
    
    def situation(self):
        if self._situation is not None:
            return self._situation
        
        if self.is_outside():
            return Outdoors(self.model())
        
        return None
    
    def set_situation(self, situation):
        self._situation = situation
    
    def resolve_situation(self, situation):
        if self._situation == situation:
            self._situation = None

    def look_around(self):
        return self.name()
        
    def events(self, avatar):
        if self.is_hidden():
            return [FindStructure(self, avatar, self)]
        else:
            return [SeeStructure(self, avatar, self)]

    def __str__(self):
        if self.name()[0].lower() in 'aeiou':
            return 'an %s' % self.name()
        else:
            return 'a %s' % self.name()



class PhoenixNest(PhoenixStructure):
    def __init__(self, model, tile, domain):
        name = 'Phoenix Nest'
        item = Feather(model, domain)
        
        PhoenixStructure.__init__(self, model, tile, domain, True, True,
                                  item, name)
        
        self._home = random.randint(0, 1)

    def is_home(self):
        return self._home
    
    def description(self):
        if self.is_home():
            return ['A majestic %s phoenix is flying above!' % self.domain()]
        else:
            return ['No phoenix seems to be near.']
        
    def fly(self):
        self._home= not self._home

    
    
class SunkenCity(PhoenixStructure):
    def __init__(self, model, tile):
        domain = 'water'
        name = 'Sunken City'
        item = Gems(model, domain, random.randint(7, 10))
        
        PhoenixStructure.__init__(self, model, tile, domain, True, True,
                                  item, name)

    def is_home(self):
        return self._home
    
    def description(self):
        return ['It seems to have sunken a long time ago.']

    
    
class TreeOfAges(PhoenixStructure):
    def __init__(self, model, tile):
        domain = 'earth'
        name = 'Tree of Ages'
        item = Gems(model, domain, random.randint(9, 12))
        
        PhoenixStructure.__init__(self, model, tile, domain, True, True,
                                  item, name)
        
    def description(self):
        return ['Very old trees are standing around you.',
                'You are at the very heart of the forest.']

    
    
class LavaForge(PhoenixStructure):
    def __init__(self, model, tile):
        domain = 'fire'
        name = 'Lava Forge'
        item = Gems(model, domain, random.randint(11, 14))
        
        PhoenixStructure.__init__(self, model, tile, domain, True, True,
                                  item, name)
        
    def description(self):
        return ['This forge was once used by artificers!',
                'Many great artifacts were made in these fires.']

    
    
class IvoryTower(PhoenixStructure):
    def __init__(self, model, tile):
        domain = 'air'
        name = 'Ivory Tower'
        item = Gems(model, domain, random.randint(13, 16))
        
        PhoenixStructure.__init__(self, model, tile, domain, True, True,
                                  item, name)
        
    def description(self):
        return ['It is full of laboratories.',
                'The mage who lived here was very powerful!']

    
    
class Hut(PhoenixStructure):
    def __init__(self, model, tile):
        PhoenixStructure.__init__(self, model, tile, None, False, True,
                                  None, 'Small Hut')
    
    def is_wooden(self):
        return True
    
    def description(self):
        return ['This is your home. You need to leave it now.']

    
    
class AlchemistsHut(PhoenixStructure):
    def __init__(self, model, tile):
        PhoenixStructure.__init__(self, model, tile, None, True, True,
                                  None, 'Alchemist\'s Hut')
    
        self._remaining = random.randint(1, 3)
    
    def remaining(self):
        return self._remaining

    def buy_potion(self):
        self._remaining -= 1
        
    def potions(self):
        if self.remaining() == 0:
            return []
        
        return [( 100,  100),
                ( 250,  500),
                ( 500, 1500),
                (1000, 4000)]
    
    def situation(self):
        return AlchemistSituation(self.model(), self.model().current_avatar(),
                                  self)
    
    def description(self):
        if self.remaining() == 0:
            return ['The hut is empty. There\'s a sign at the door:',
                    '"I\'m out gathering herbs."',
                    '"Back an hour or so after rebirth."']
        
        return ['"Would you like me to brew you a potion?"',
                '"I have ingredients for %d of them left."' % self.remaining()]

    
    
class BlackMarket(PhoenixStructure):
    def __init__(self, model, tile):
        PhoenixStructure.__init__(self, model, tile, None, True, True,
                                  None, 'Black Market')
    
        self._buys  = []
        self._sells = []
        self._value = random.randint(3, 5) * 10
    
    def value(self):
        return self._value
    
    def is_dark_place(self):
        return True
    
    def domains(self):
        return ['air', 'fire', 'earth', 'water'] + self.tile().terrain.domains()
    
    def choose_domain(self):
        return random.choice(self.domains())
    
    def create_buy(self):
        avatar = self.model().current_avatar()
        num    = random.randint(1, 5)
        price  = num * random.randint(7, 12) * self.value()
        
        return BlackMarketTrade(avatar, self, False, 
                                num, self.choose_domain(), price)
    
    def create_sell(self):
        avatar = self.model().current_avatar()
        num    = random.randint(1, 3)
        price  = num * random.randint(2, 5) * self.value()
        
        return BlackMarketTrade(avatar, self, True, 
                                num, self.choose_domain(), price)
    
    def remove_trade(self, trade):
        if trade in self._buys:
            self._buys.remove(trade)
            
        if trade in self._sells:
            self._sells.remove(trade)
        
    def trades(self):
        while len(self._buys) < 2 or str(self._buys[0]) == str(self._buys[1]):
            if len(self._buys) < 2:
                self._buys.append(self.create_buy())
            else:
                self._buys[1] = self.create_buy()
        
        while len(self._sells) < 2 or str(self._sells[0]) == str(self._sells[1]):
            if len(self._sells) < 2:
                self._sells.append(self.create_sell())
            else:
                self._sells[1] = self.create_sell()
                
        return self._buys + self._sells
    
    def situation(self):
        return BlackMarketSituation(self.model(), self.model().current_avatar(),
                                    self)
    
    def description(self):
        return ['A whispering voice asks you for your business.']

    
    
class Town(PhoenixStructure):
    def __init__(self, model, tile):
        PhoenixStructure.__init__(self, model, tile, None, False, True,
                                  None, 'Town')
    
        self._bought = 0
    
    def bought(self):
        return self._bought

    def buy_map(self):
        self._bought += 1
        
    def maps(self):
        return [('Map', [AlchemistsHut, PhoenixNest, Mine], 
                 1000 + 500 * self.bought()),
                ('Shadow Map', [BlackMarket],
                 1500 + 500 * self.bought()),
                ('Secret Map', [WishingWell, FountainOfYouth,
                                DragonCave, CrystalPlatform], 
                 2000 + 500 * self.bought())]
    
    def situation(self):
        return TownSituation(self.model(), self.model().current_avatar(), self)
    
    def description(self):
        return ['Preperations for a festival are going on.',
                'Only very few stores are open.']

    

class SituationStructure(PhoenixStructure):
    def __init__(self, model, tile, domain, hidden, name):
        PhoenixStructure.__init__(self, model, tile, domain, hidden, True,
                                  None, name)
    
    def avatar(self):
        return self.model().current_avatar()
    
    def map(self):
        return self.model().current_map()
    
    def situation(self):
        return StructureSituation(self)
    
    def actions(self):
        raise 'Not implemented'
    
    def description(self):
        raise 'Not implemented'
    
    
    
class WishingWell(SituationStructure):
    def __init__(self, model, tile):
        SituationStructure.__init__(self, model, tile, 'water', True,
                                    'Wishing Well')
    
        self._wished = False
    
    def wished(self):
        return self._wished

    def wish(self):
        self._wished = True
        
    def actions(self):
        if self.wished():
            return []
        
        wish   = ['Your wish fullfils itself!']
        wished = [lambda: self.wish()]
        
        return [('Wish for Health', 1,
                 wish + ['You gain 1400 health.'],
                 wished + [lambda: self.avatar().heal(1400)]),
                ('Wish for Money', 1,
                 wish + ['You gain 7000 gold.'],
                 wished + [lambda: self.avatar().add_gold(7000)]),
                ('Wish for Power', 1,
                 wish + ['You gain 9 of every gem.'],
                 wished + [lambda: self.avatar().add_gems('all', 9)])]
    
    def description(self):
        if self.wished():
            return []
        
        return ['You may wish for something!']
    
    
    
class DragonCave(SituationStructure):
    def __init__(self, model, tile):
        SituationStructure.__init__(self, model, tile, 'fire', True,
                                    'Dragon Cave')
    
        self._dragon = False
        self._gold   = 10000
        
    def dragon(self):
        return self._dragon
    
    def gold(self):
        return self._gold

    def slay(self):
        self._dragon = True
    
    def comes_back(self):
        return not self.dragon() and random.randint(0, 3) == 0
    
    def take(self):
        if self._gold > 0:
            self._gold -= 1000
            self.avatar().add_gold(1000)
        
    def actions(self):
        if self.gold() == 0:
            return []
        
        if self.comes_back():
            text    = ['You suddenly hear a roar behind you.',
                       'The entrance is blocked by a big dragon!',
                       'The dragon hurls fire for 1000 damage at you!']
            actions = [lambda: self.slay(),
                       lambda: self.avatar().hurt(1000)]
            
            if self.avatar().health() > 1000:
                text += ['In an epic battle, you slay it!']
        else:
            if self.dragon():
                text = ['You take 1000 gold.',
                        'There is %d gold left.' % (self.gold() - 1000)]
            else:
                text = ['You take 1000 gold. You suddenly shudder.',
                        'You look at the entrance.',
                        'Everything is quiet.',
                        'There is %d gold left.' % (self.gold() - 1000)]
            actions = [lambda: self.take()]
            
        return [('Take some Gold', None, text, actions)]
    
    def description(self):
        if self.dragon():
            dragon = ['There is the body of a dragon here.']
        else:
            dragon = ['The dragon is not here... at this moment!']
        
        if self.gold():
            return ['There is a pile of %d gold here!' % self.gold()] + dragon
        else:
            return ['The cave is empty.']
    
    
    
class CrystalPlatform(SituationStructure):
    def __init__(self, model, tile):
        SituationStructure.__init__(self, model, tile, 'air', True,
                                    'Crystal Platform')
    
        self._used = False
        
    def used(self):
        return self._used
    
    def use(self):
        self._used = True
        
    def actions(self):
        if self.used():
            return []

        return [('Step on the Platform', None, 
                ['You step onto the platform. It begins spinning!',
                 'You feel the magic in you flow into one point.',
                 'You suddenly hover over the platform!'],
                [lambda: self.use(),
                 lambda: self.avatar().forget_skills(),
                 lambda: self.avatar().add_skill(Floating())])]
        
    def description(self):
        if self.used():
            return []
        
        return ['It is hovering slightly above ground!',
                'You hear a voice in your mind: "Are you lost?"',
                '"Then I can help you. But the price is high!"']
    
    
class RainbowFoundation(SituationStructure):
    def __init__(self, model, tile):
        SituationStructure.__init__(self, model, tile, None, False,
                                    'Rainbow Foundation')
        
    def can_win(self):
        return reduce(lambda x, domain: (x and 
                                         self.avatar().feathers(domain) > 0),
                                         ['water', 'earth', 'fire', 'air'])
        
    def actions(self):
        if not self.can_win():
            return []
        
        return [('Invoke the Rainbow', None, 
                 ['You put the feathers on the altar.',
                  'They start glowing and vanish!',
                  'Suddenly, the rainbow outline is filled!',
                  'You are standing in front of a rainbow bridge.'], 
                 [lambda: self.model().win()])]
    
    def description(self):
        text = []
        
        if self.can_win():
            text = ['Suddenly, the flimmering stops. An altar appears!']
            
        return ['You see the outlines of a rainbow in the air.',
                'You see something flimmering at the base.'] + text
    
    
class Mine(SituationStructure):
    def __init__(self, model, tile):
        SituationStructure.__init__(self, model, tile, None, True,
                                    'Mine')
    
        self._domains = {'water': random.randint(1, 20),
                         'earth': random.randint(1, 20),
                         'fire' : random.randint(1, 20),
                         'air'  : random.randint(1, 20)}
        
        self._gems  = 7
        self._order = self._domains.keys()
        self._order.sort(lambda a, b: cmp(self._domains[b], self._domains[a]))
    
    def gems(self):
        return self._gems
    
    def gem(self):
        if random.randint(0, self._gems * 2) == 0:
            return None
        
        domain = self.random_domain()
        
        if domain is None:
            return None
        
        return Gem(self, domain)
    
    def mining(self, chance):
        domain = self.random_domain()
        
        if domain is None:
            return None
        
        if random.randint(1, 100) <= chance:
            return Gems(self, domain, random.randint(2, 4))
        else:
            return Gem(self, domain)

    def take_gem(self):
        self._gems -= 1
        
    def order(self):
        return self._order
    
    def random_domain(self):
        chance = random.randint(1, 100)
        
        for domain in self._domains.keys():
            if chance <= self.chance(domain):
                return domain
            
            chance -= 25
        
        return None
    
    def inspection(self):
        result = []
        
        for domain in self.order():
            chance = self.chance(domain)
            name   = Gems(self, domain, 2).name()
            
            if chance > 16:
                text = 'plenty of'
            elif chance > 12:
                text = 'a lot of'
            elif chance > 8:
                text = 'a number of'
            elif chance > 4:
                text = 'a few'
            else:
                text = 'hardly any'
            
            result += ['It should yield %s %s.' % (text, name)]
        
        return result
    
    def chance(self, domain):
        return self._domains[domain]
        
    def actions(self):
        gem = self.gem()
        
        if gem is not None:
            s_text = ['You find %s!' % str(gem)]
            find   = [lambda: gem.found_by(self.avatar()),
                      lambda: self.take_gem()]
        else:
            s_text = ['You find nothing.']
            find   = []
            
        inspect = ('Inspect the Mine', None, self.inspection(), [])
        search  = ('Search the Mine',  None, ['You search the mine.'] + s_text, 
                   find)
        
        if self.avatar().can_mine():
            result = self.mining(self.avatar().mining())
            
            if result is None or random.randint(0, 2) == 0:
                m_text = ['You find nothing.']
                m_find = []
            else:
                m_text = ['You find %s!' % str(result)]
                m_find = [lambda: result.found_by(self.avatar())]
              
            mine = ('Dig',  None, ['You dig for gems.'] + m_text, 
                    m_find)
            
            return [inspect, search, mine]
        
        return [inspect, search]
    
    def description(self):
        return ['It has been abandoned years ago.']

    
    
class FountainOfYouth(SituationStructure):
    def __init__(self, model, tile):
        SituationStructure.__init__(self, model, tile, 'earth', True,
                                    'Fountain of Youth')
    
        self._done = False
    
    def is_done(self):
        return self._done

    def done(self):
        self._done = True
        
    def actions(self):
        if self.is_done():
            return []
        
        text = ['You are rejuvenated!']
        done = [lambda: self.done()]
        gems = self.avatar().gems('water') * self.avatar().gems('earth')
        heal = int(math.sqrt(gems) * 2 + random.randint(-10, 20)) * 50

        if heal < 500:
            heal = 100
            
        if heal > 4000:
            heal = 4000
        
        return [('Drink the Water', None,
                 ['You drink from the fountain.',
                  'The water tastes like every other water.'] + text + 
                 ['You gain 1000 health.'],
                 done + [lambda: self.avatar().heal(2000)]),
                ('Take a Bath', None,
                 ['You take a bath in the fountain.',
                  'You feel your power mixing with the water!'] + text + 
                 ['You gain %d health.' % heal],
                 done + [lambda: self.avatar().heal(heal)])]
    
    def description(self):
        if self.is_done():
            return []
        
        return ['The water of life is flowing from it.']
