'''Terrain in the Phoenix game.'''

import random

from concepts.map import Terrain

from phoenix.events import *
from phoenix.items import *
from phoenix.monsters import *
from phoenix.situations import *
from phoenix.structures import RainbowFoundation



class PhoenixTerrain(Terrain):
    def __init__(self, model, tile, name):
        Terrain.__init__(self, model, tile, name)
        
        self._touched = 0
        self._situation = None
    
    def special_land(self):
        return False
    
    def scorch(self, degree, all = False):
        pass
    
    def touched(self):
        return self._touched
    
    def monster(self):
        if random.randint(0, 20) < self.touched():
           return None

        return random.choice(self.inhabitants())()
    
    def gem(self):
        deposits = int(7 * self.model().current_avatar().survey())
        
        if random.randint(0, deposits) < self.touched():
           return PreciousStone(self.model(), deposits * 700 /
                                (10 + self.touched()))

        return Gem(self.model(), random.choice(self.domains()))
    
    def fixed_events(self, avatar):
        if self.tile().is_built():
            if self.tile().structure().is_hidden():
                return [FindStructure(self, avatar, self.tile().structure())]
            else:
                return [SeeStructure(self, avatar, self.tile().structure())]
        
        if self.model().current_avatar().conjuring():
            return [ConjureItem(self, avatar, Gem(self.model(), 'air'))]
        
        return []
    
    def closest_nest(self):
        return self.tile().closest(lambda tile: isinstance(tile.structure(),
                                                           PhoenixNest))
        
    def events(self, avatar):
        self._touched += 1
        
        return self.create_events(avatar)
        
    def domains(self):
        raise []
    
    def create_events(self, avatar):
        raise 'Not implemented'
    
    def inhabitants(self):
        raise 'Not implemented'
    
    def situation(self):
        if self._situation is None:
            return Outdoors(self.model())
        
        return self._situation
    
    def set_situation(self, situation):
        self._situation = situation
    
    def resolve_situation(self, situation):
        if self._situation == situation:
            self._situation = None
    
    def life(self):
        return 0
    
    def water_depth(self):
        return 0
    
    def ambush(self, avatar, monster):
        peace  = avatar.is_peace()
        events = [Ambush(self, avatar, monster, peace)]
        
        if peace:
            events += [Hug(avatar, monster)]
        elif avatar.fiery_aura() >= monster.strength():
            events += [Ashes(avatar, monster)]
            
        return events


    
    
class Plains(PhoenixTerrain):
    def __init__(self, model, tile, water = 30):
        PhoenixTerrain.__init__(self, model, tile, 'Plains')
        
        self._life = 55 + water + random.randint(-10, 10)
        
        if self._life > 100:
            self._life = 100
        
    def life(self):
        return self._life
    
    def scorch(self, degree, all = False):
        self._life -= degree * 10
        
        if self._life < 40 * (1 - all):
            self._life = 40 * (1 - all)
    
    def inhabitants(self):
        people = [BANDIT, GOBLIN, ORC, TROLL]
        num    = (self.life() - 60) / 10
        
        if num > 0:
            people += [BANDIT] * num
        
        return people
    
    def domains(self):
        domains = []
        
        if self.life() < 60:
            domains += ['fire']
        else:
            domains += ['water']
        
        if self.life() < 80:
            domains += ['earth']
        else:
            domains += ['air']
            
        return domains
        
    def create_events(self, avatar):
        events = self.fixed_events(avatar)
        
        if events:
            return events
        
        chance = random.randint(0, 100)
        dmg    = (60 - self.life() + 1) / 2
        heal   = (self.life() - 77) / 3
        
        if dmg > 0:
            if dmg > 6:
                if chance < 20:
                    return [Thirst(self, avatar, dmg * 6, dmg * 10)]
                
                if chance < 50:
                    return [Heat(self, avatar, dmg * 3, dmg * 5)]
                    
                return [Flavor(self, avatar, ['This land has been scorched.'])]
            
            if chance < dmg * 4:
                return [FindItem(self, avatar, self.gem())]
                
            if chance < 20:
                return [Thirst(self, avatar, dmg * 10, dmg * 20)]

            if chance < 25:
                return [Trap(self, avatar, 50, 100)]
            
            if chance < 50:
                return [Heat(self, avatar, dmg * 3, dmg * 5)]
        
            if chance < 60:
                return [PhoenixFlight(self, avatar, self.closest_nest())]
        
            if chance < 70:
                monster = self.monster()
                
                if monster is not None:
                    return self.ambush(avatar, monster)
                
            return [Flavor(self, avatar, ['Barely any grass grows here.',
                                          'You see only dying plants.',
                                          'The ground is barren.',
                                          'You see cracks in the ground.',
                                          'Heat has dried out the ground.',
                                          'The sun burns down mercilessly.']
                                          [dmg - 1:dmg + 1])]
            
        if heal > 0:
            if heal > 7:
                heal = 7
                
            if chance < 50:
                return [Tranquility(self, avatar, heal * 5, heal * 7)]
        
            if chance < 80:
                return [PhoenixFlight(self, avatar, self.closest_nest())]
            
            if chance < 85:
                return [FindItem(self, avatar, self.gem())]
             
            return [Flavor(self, avatar, ['Grass grows tall here.',
                                          'You see a beautiful meadow.',
                                          'Some very old trees grow here.',
                                          'You see a flock of birds in the sky.',
                                          'Animals are playing in the fields.',
                                          'You see all kinds of species playing together.',
                                          'You get a deep feeling of peace.']
                                          [heal - 1:heal + 1])]
                    
        if chance < 10:
            return [FindItem(self, avatar, self.gem())]
        
        if chance < 30:
            return [PhoenixFlight(self, avatar, self.closest_nest())]
        
        if chance < 35:
            return [Trap(self, avatar, 50, 100)]
      
        if chance < 50:
            gold = (random.randint(5, 10) - self.touched()) * 100
            
            if gold < 100:
                gold = 50
            
            if random.randint(0, 1):
                open = ('Open', None, ['You open the chest. It wasn\'t trapped.',
                                       'It contains %d gold.' % gold],
                        [lambda: avatar.add_gold(gold)])
            else:
                damage = random.randint(50, 250)
                res    = avatar.resistance(damage)
                
                if damage > 150 and random.randint(0, 2):
                    damage -= 100
                    
                open = ('Open', None, ['You open the chest. It was trapped!',
                                       'You suffer %d damage.' % (damage - res),
                                       'It contains %d gold.' % gold],
                        [lambda: avatar.damage(damage),
                         lambda: avatar.add_gold(gold)])
            
            leave = ('Leave', None, ['You leave the chest.'], [])
            
            return [SituationEvent(self, ['You find a treasure chest!'],
                                   [open, leave], False)]
            
        if chance < 70:
            monster = self.monster()
            
            if monster is not None:
                return self.ambush(avatar, monster)
        
        return [Flavor(self, avatar, ['You see a meadow.',
                                      'You see grass all around you.',
                                      'You see a small road.',
                                      'You hear a bird singing.',
                                      'A raven flies high above you.',
                                      'You see a small creek.',
                                      'You see a grove of trees.',
                                      'You exchange pleasantries with a traveller.',
                                      'You see kids playing hide and seek.'])]

    def look_around(self):
        if self.life() >= 80:
            return 'Idyllic Plains'

        if self.life() >= 60:
            return 'Plains'
        
        return 'Barren Plains'
    
    
    
class Lake(PhoenixTerrain):
    def __init__(self, model, tile, depth = None):
        PhoenixTerrain.__init__(self, model, tile, 'Lake')
        
        if depth is None:
            self._depth = random.randint(1, 100)
        else:
            self._depth = depth
            
        if self._depth < 1:
            self._depth = 1
            
        if self._depth > 100:
            self._depth = 100
        
    def life(self):
        return 85 - self.water_depth() / 2
    
    def scorch(self, degree, all = False):
        self._depth -= degree * 20
        
        if self._depth <= 0:
            self.tile().become(Plains)
    
    def inhabitants(self):
        return [MERMAID, MERMAN]
    
    def domains(self):
        return ['water']
        
    def create_events(self, avatar):
        events = self.fixed_events(avatar)
        
        if events:
            return events

        chance = random.randint(0, 100)

        if chance < 20:
            gem = self.gem()
            
            if isinstance(gem, Gem):
                return [FindItem(self, avatar, gem)]
        
        if chance < 30:
            return [PhoenixFlight(self, avatar, self.closest_nest())]

        if chance < 40:
            monster = self.monster()
            
            if monster is not None:
                return self.ambush(avatar, monster)
        
        if chance < self.water_depth() * 5 and not avatar.is_aquatic():
            return [Drown(self, avatar, 
                          self.water_depth(), int(self.water_depth() * 2.5))]
            
        return [Flavor(self, avatar, ['You see fishies swimming around you.',
                                      'The water is quiet.'])]

    def look_around(self):
        if self.water_depth() >= 40:
            return 'Lake'

        return 'Shallow Lake'
    
    def water_depth(self):
        return self._depth
    
    
    
class Forest(PhoenixTerrain):
    def __init__(self, model, tile, deep = False):
        PhoenixTerrain.__init__(self, model, tile, 'Forest')
        
        if deep:
            self._life = random.randint(85, 100)
        else:
            self._life = random.randint(65, 75)
            
    def life(self):
        return self._life
    
    def scorch(self, degree, all = False):
        self._life -= degree * 20
        
        if self._life <= 50:
            self.tile().become(Plains)
    
    def inhabitants(self):
        return [ELF, ELF, ELF, ELF, TREANT, DRUID, DRUID, DRYAD]
    
    def domains(self):
        return ['earth', 'water']
        
    def create_events(self, avatar):
        events = self.fixed_events(avatar)
        
        if events:
            return events

        chance = random.randint(0, 100)

        if chance < 10:
            gem = self.gem()
            
            if isinstance(gem, Gem):
                return [FindItem(self, avatar, gem)]
        
        if chance < 25:
            return [PhoenixFlight(self, avatar, self.closest_nest())]
        
        if chance < 25 + avatar.forest_riches():
            return [TextEvent(['You hug a tree! You see a sparkle at its roots.']),
                    FindItem(self, avatar, Gem(self.model(), 'earth'))]
                 
        if chance < 80:
            monster = self.monster()
            
            if monster is not None:
                return self.ambush(avatar, monster)
                    
        return [Flavor(self, avatar, ['You see a squirrel climbing a tree.',
                                      'You barely find a way through the trees.',
                                      'A deer notices you and flees.',
                                      'You see an old oak.'])]

    def look_around(self):
        if self.life() >= 80:
            return 'Deep Forest'
        
        if self.life() >= 60:
            return 'Forest'
        
        return 'Barren Forest'
    
    def can_move_in(self, avatar):
        if self.life() < 80:
            return True

        return (avatar.is_forest_walking() or avatar.is_floating())
    
    def special_land(self):
        return True
    
    
    
class Hills(PhoenixTerrain):
    def __init__(self, model, tile):
        PhoenixTerrain.__init__(self, model, tile, 'Hills')

        self._life = random.randint(45, 65)
            
    def life(self):
        return self._life
    
    def scorch(self, degree, all = False):
        self._life -= degree * 5
        
        if self._life <= 40:
            if all:
                self.tile().become(Plains)
            else:
                self._life = 41
                    
    
    def inhabitants(self):
        if self.life() >= 60:
            return [ROBBER, ROBBER, HILL_ORC]
        else:
            return [ROBBER, HILL_ORC, HILL_ORC, OGRE]
    
    def domains(self):
        return ['earth', 'fire']
        
    def create_events(self, avatar):
        events = self.fixed_events(avatar)
        
        if events:
            return events

        chance = random.randint(0, 100)

        if chance < 20:
            gem = self.gem()
            
            if isinstance(gem, Gem):
                return [FindItem(self, avatar, gem)]
        
        if chance < 25:
            return [PhoenixFlight(self, avatar, self.closest_nest())]

        if chance < 30:
            return [Trap(self, avatar, 50, 100)]
      
        if chance < 40:
            gold = (random.randint(5, 10) - self.touched()) * 100
            
            if gold < 100:
                gold = 50
            
            if random.randint(0, 1):
                open = ('Open', None, ['You open the chest. It wasn\'t trapped.',
                                       'It contains %d gold.' % gold],
                        [lambda: avatar.add_gold(gold)])
            else:
                damage = random.randint(50, 250)
                res    = avatar.resistance(damage)
                
                if damage > 150 and random.randint(0, 2):
                    damage -= 100
                    
                open = ('Open', None, ['You open the chest. It was trapped!',
                                       'You suffer %d damage.' % (damage - res),
                                       'It contains %d gold.' % gold],
                        [lambda: avatar.damage(damage),
                         lambda: avatar.add_gold(gold)])
            
            leave = ('Leave', None, ['You leave the chest.'], [])
            
            return [SituationEvent(self, ['You find a treasure chest!'],
                                   [open, leave], False)]
        
        if chance < 80:
            monster = self.monster()
            
            if monster is not None:
                return self.ambush(avatar, monster)
                    
        return [Flavor(self, avatar, ['A small river runs at the foot of this hill.'])]

    def look_around(self):
        if self.life() >= 60:
            return 'Grassy Hills'
        
        return 'Hills'
    
    def special_land(self):
        return True
    
    
    
class Mountain(PhoenixTerrain):
    def __init__(self, model, tile):
        PhoenixTerrain.__init__(self, model, tile, 'Mountain')

        self._life = 50
            
    def life(self):
        return self._life
    
    def scorch(self, degree, all = False):
        self._life -= degree * 5
        
        if self._life <= 40:
            if all:
                self.tile().become(Hills)
            else:
                self._life = 41
                    
    
    def inhabitants(self):
        return [YETI, MOUNTAIN_TROLL, MOUNTAIN_TROLL, MOUNTAIN_TROLL]
    
    def domains(self):
        return ['earth', 'air']
        
    def create_events(self, avatar):
        events = self.fixed_events(avatar)
        
        if events:
            return events

        chance = random.randint(0, 100)

        if chance < 10:
            gem = self.gem()
            
            if isinstance(gem, Gem):
                return [FindItem(self, avatar, gem)]
        
        if chance < 15:
            return [PhoenixFlight(self, avatar, self.closest_nest())]

        if chance < 25:
            return [Trap(self, avatar, 75, 150)]
        
        if chance < 50:
            monster = self.monster()
            
            if monster is not None:
                return self.ambush(avatar, monster)
        
        if chance < 50 + avatar.mountain_riches():
            gold = 10000 / (10 + self.touched())
            gems = [PreciousStone(self, random.randint(gold / 2, gold * 2))
                    for nr in range(random.randint(2, 3))]
            
            return [DirectEvent(['You find a cave! In it you find:'] +
                                ['- %s' % str(gem) for gem in gems],
                                [(lambda: gem.found_by(avatar))
                                 for gem in gems])]
            
        return [Flavor(self, avatar, ['You have a wonderful view.',
                                      'You find a beautiful cave.'])]

    def look_around(self):
        return 'Mountain'
    
    def special_land(self):
        return True

    def can_move_in(self, avatar):
        return (avatar.is_climbing() or avatar.is_floating())
    
    
    
class Glacier(PhoenixTerrain):
    def __init__(self, model, tile):
        PhoenixTerrain.__init__(self, model, tile, 'Glacier')

        self._life = 50
            
    def life(self):
        return self._life
    
    def scorch(self, degree, all = False):
        self._life -= degree * 5
        
        if self._life <= 40:
            if all:
                self.tile().become(Mountain)
            else:
                self._life = 41
                    
    
    def inhabitants(self):
        return [YETI, YETI, YETI, MOUNTAIN_TROLL]
    
    def domains(self):
        return ['air']
        
    def create_events(self, avatar):
        events = self.fixed_events(avatar)
        
        if events:
            return events

        chance = random.randint(0, 100)

        if chance < 10:
            gem = self.gem()
            
            if isinstance(gem, Gem):
                return [FindItem(self, avatar, gem)]
        
        if chance < 50:
            monster = self.monster()
            
            if monster is not None:
                return self.ambush(avatar, monster)
                    
        return [Flavor(self, avatar, ['Brrr. It is cold up here!',
                                      'You stumble and fall into the snow.',
                                      'You have a wonderful view.'])]

    def look_around(self):
        return 'Glacier'
    
    def special_land(self):
        return True

    def can_move_in(self, avatar):
        return (avatar.is_climbing() or avatar.is_floating())
    
    
    
class Lava(PhoenixTerrain):
    def __init__(self, model, tile):
        PhoenixTerrain.__init__(self, model, tile, 'Lava')

        self._life = 5
        
    def life(self):
        return self._life
    
    def scorch(self, degree, all = False):
        if all:
            self._life -= 1
        
    def domains(self):
        return ['fire']
          
    def create_events(self, avatar):
        events = self.fixed_events(avatar)
        heat   = [LavaHeat(self, avatar, 100, 200)]
        
        if events:
            return events

        chance = random.randint(0, 100)
        
        if avatar.forging() and chance < 50:
            if avatar.forging() == 1:
                item = Gem(self, 'fire')
            else:
                item = Gems(self, 'fire', avatar.forging())
                
            return heat + [DirectEvent(['You forge %s!' % item,
                                        'Your power in the domain of fire increases.'],
                                       [lambda: item.found_by(avatar)])]
        
        if avatar.minting() and chance >= 50:
            gold = avatar.minting()
            
            return heat + [DirectEvent(['You mint %d gold!' % gold],
                                       [lambda: avatar.add_gold(gold)])]

        if chance < 5:
            gem = self.gem()
            
            if isinstance(gem, Gem):
                return heat + [FindItem(self, avatar, gem)]
        
        if chance < 95:
            return heat
        
        return [Flavor(self, avatar, ['You manage to find a relative cool place.'])]

    def look_around(self):
        return 'Lava'
    
    def special_land(self):
        return True
    
    
    
class RainbowMountain(PhoenixTerrain):
    def __init__(self, model, tile):
        PhoenixTerrain.__init__(self, model, tile, 'Rainbow Mountain')
        
        tile.build(RainbowFoundation)
        
        self._life = 100
        
    def life(self):
        return self._life
    
    def look_around(self):
        return 'Rainbow Mountain'
    
    def special_land(self):
        return True
    
    
    
class Apocalyptic(PhoenixTerrain):
    def __init__(self, model, tile):
        PhoenixTerrain.__init__(self, model, tile, 'Apocalyptic Place')
                
    def create_events(self, avatar):
        return [Death(self, avatar, True)]
    
    def look_around(self):
        return 'Apocalyptic Place'
    
    def special_land(self):
        return True
    
